from pathlib import Path

import numpy as np

from .utils import StrictBaseModel


class LUT(StrictBaseModel):
    """
    A class for defining look up tables
    """

    path: Path | str
    """
    The path at which we will read/write the look up table
    """

    @property
    def name(self) -> str:
        return self.file.stem

    @property
    def file(self) -> Path:
        if isinstance(self.path, str):
            return (
                Path(self.path)
                if self.path.endswith(".npz")
                else Path(f"{self.path}.npz")
            )
        elif isinstance(self.path, Path):
            return self.path

    def drop(self):
        if self.file.exists():
            self.file.unlink()

    def save(self, lut, edges=None):
        np.savez_compressed(self.file, fragment_segment_lut=lut, edges=edges)

    def load(self) -> np.ndarray:
        return np.load(self.file)["fragment_segment_lut"]



import logging
from contextlib import contextmanager
from itertools import chain
from typing import Annotated, Callable, Generator, Literal

import networkx as nx
import numpy as np
import scipy.ndimage
from daisy import Block
from funlib.geometry import Coordinate, Roi
from funlib.math import inv_cantor_number
from funlib.persistence.arrays import Array
from funlib.persistence.graphs.graph_database import GraphDataBase
from pydantic import Field

from ..datasets import Affs, Dataset, Labels
from ..dbs import PostgreSQL, SQLite
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask

logger = logging.getLogger(__file__)


class AffAgglom(BlockwiseTask):
    """
    A blockwise task that computes edges between supervoxels
    sharing affinity edges and stores statistics such as the
    mean affinity value between the supervoxels.
    """

    task_type: Literal["aff-agglom"] = "aff-agglom"
    db: Annotated[
        PostgreSQL | SQLite,
        Field(discriminator="db_type"),
    ]
    """
    The database containing nodes associated with the fragment supervoxels
    to which we will add the supervoxel edge affinity statistics
    """
    affs_data: Affs
    """
    The affs to read for creating supervoxel affinity statistics
    """
    frags_data: Labels
    """
    The labels array containing the supervoxels for computing affinity statistics
    """
    block_size: PydanticCoordinate
    """
    The blocksize within which to compute supervoxel affinity statistics
    """
    context: PydanticCoordinate
    """
    The amount of context to use when computing supervoxel affinity statistics
    """
    scores: dict[str, list[PydanticCoordinate]]
    """
    A dictionary of score names and their respective neighborhoods.
    This allows us to compute the affinity statistics in subgroups of the
    neighborhood. For example if you wanted to compute the mean affinity
    between fragments in x and y separately from z, you would provide
    a dictionary like so:

        scores = {
            "xy": [(1, 0, 0), (0, 1, 0)],
            "z": [(0, 0, 1)]
        }

    """
    fit: Literal["shrink"] = "shrink"
    """
    The boundary behavior for our daisy task.
    """
    read_write_conflict: Literal[False] = False
    """
    We don't have read write conflicts in this task and can compute
    every block independently in an arbitrary order.
    """

    @property
    def task_name(self) -> str:
        return f"{self.db.id}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.frags_data.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.frags_data.array("r").voxel_size

    @property
    def context_size(self) -> Coordinate:
        return self.context * self.frags_data.array("r").voxel_size

    @property
    def output_datasets(self) -> list[Dataset]:
        return []

    def drop_artifacts(self):
        self.db.drop_edges()

    def agglomerate(self, affs: np.ndarray, frags: np.ndarray, rag: nx.Graph):
        fragment_ids = [int(x) for x in np.unique(frags) if x != 0]
        num_frags = len(fragment_ids)
        frag_mapping = {
            old: seq for seq, old in zip(range(1, num_frags + 1), fragment_ids)
        }
        rev_mapping = {v: k for k, v in frag_mapping.items()}
        for old, seq in frag_mapping.items():
            frags[frags == old] = seq

        if len(fragment_ids) == 0:
            return

        def count_affs(
            fragments: np.ndarray, affinities: np.ndarray, offset: Coordinate
        ) -> dict[int, tuple[float, float]]:
            base_frags = frags[
                tuple(
                    slice(-m if m < 0 else None, -m if m > 0 else None) for m in offset
                )
            ]
            base_affinities = affinities[
                tuple(
                    slice(-m if m < 0 else None, -m if m > 0 else None) for m in offset
                )
            ]
            offset_frags = fragments[
                tuple(slice(m if m > 0 else None, m if m < 0 else None) for m in offset)
            ]

            mask = (offset_frags != base_frags) * (offset_frags > 0) * (base_frags > 0)

            # cantor pairing function
            # 1/2 (k1 + k2)(k1 + k2 + 1) + k2
            k1, k2 = (
                np.min(
                    [
                        offset_frags,
                        base_frags,
                    ],
                    axis=0,
                ),
                np.max(
                    [
                        offset_frags,
                        base_frags,
                    ],
                    axis=0,
                ),
            )
            cantor_pairings = ((k1 + k2) * (k1 + k2 + 1) / 2 + k2) * mask
            cantor_ids = np.array([x for x in np.unique(cantor_pairings) if x != 0])
            scores = scipy.ndimage.mean(
                base_affinities,
                cantor_pairings,
                cantor_ids,
            )
            counts = scipy.ndimage.sum_labels(
                mask,
                cantor_pairings,
                cantor_ids,
            )
            mapping = {
                cantor_id: (mean_score, count)
                for cantor_id, mean_score, count in zip(cantor_ids, scores, counts)
            }
            return mapping

        neighborhood_affs: dict[Coordinate, dict[int, tuple[float, float]]] = {}

        # affs_data.neighborhood cannot be None, assert called to make mypy happy
        assert self.affs_data.neighborhood is not None

        for offset_affs, offset in zip(affs, self.affs_data.neighborhood):
            neighborhood_affs[offset] = count_affs(frags, offset_affs, offset)

        for score_name, score_neighborhood in self.scores.items():
            offset_counts = [neighborhood_affs[offset] for offset in score_neighborhood]
            cantor_ids = set(chain(*[x.keys() for x in offset_counts]))
            for cantor_id in cantor_ids:
                key_counts = [x.get(cantor_id, (1.0, 0.0)) for x in offset_counts]
                total_count = sum([count[1] for count in key_counts])
                if total_count > 0:
                    u, v = inv_cantor_number(cantor_id, dims=2)
                    rag.add_edge(
                        rev_mapping[u],
                        rev_mapping[v],
                        **{
                            score_name: sum(
                                [
                                    count[0] * count[1] / total_count
                                    for count in key_counts
                                ]
                            )
                        },
                    )

    def agglomerate_in_block(
        self, block: Block, affs: Array, frags: Array, rag_provider: GraphDataBase
    ):
        frags_data = frags.to_ndarray(block.read_roi, fill_value=0)
        rag = rag_provider[block.read_roi]

        affs_data = affs.to_ndarray(block.read_roi, fill_value=0)

        if affs_data.dtype == np.uint8:
            affs_data = affs_data.astype(np.float32) / 255.0

        self.agglomerate(
            affs_data,
            frags_data,
            rag,
        )

        rag_provider.write_graph(rag, block.write_roi, write_nodes=False)

    def init(self) -> None:
        self.db.init()

    @contextmanager
    def process_block_func(self) -> Generator[Callable, None, None]:
        affs = self.affs_data.array("r")
        frags = self.frags_data.array("r")
        rag_provider = self.db.open("r+")

        def process_block(block) -> None:
            self.agglomerate_in_block(
                block,
                affs,
                frags,
                rag_provider,
            )

        yield process_block



import numba
import numpy as np


def seg_to_affgraph(seg, nhood):
    nhood = np.array(nhood)

    # constructs an affinity graph from a segmentation
    # assume affinity graph is represented as:
    # shape = (e, z, y, x)
    # nhood.shape = (edges, 3)
    shape = seg.shape
    nEdge = nhood.shape[0]
    dims = nhood.shape[1]
    aff = np.zeros((nEdge,) + shape, dtype=np.int32)

    if dims == 2:
        for e in range(nEdge):
            aff[
                e,
                max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
            ] = (
                (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                    ]
                    == seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                    ]
                )
                * (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                    ]
                    > 0
                )
                * (
                    seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                    ]
                    > 0
                )
            )

    elif dims == 3:
        for e in range(nEdge):
            aff[
                e,
                max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                max(0, -nhood[e, 2]) : min(shape[2], shape[2] - nhood[e, 2]),
            ] = (
                (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                        max(0, -nhood[e, 2]) : min(shape[2], shape[2] - nhood[e, 2]),
                    ]
                    == seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                        max(0, nhood[e, 2]) : min(shape[2], shape[2] + nhood[e, 2]),
                    ]
                )
                * (
                    seg[
                        max(0, -nhood[e, 0]) : min(shape[0], shape[0] - nhood[e, 0]),
                        max(0, -nhood[e, 1]) : min(shape[1], shape[1] - nhood[e, 1]),
                        max(0, -nhood[e, 2]) : min(shape[2], shape[2] - nhood[e, 2]),
                    ]
                    > 0
                )
                * (
                    seg[
                        max(0, nhood[e, 0]) : min(shape[0], shape[0] + nhood[e, 0]),
                        max(0, nhood[e, 1]) : min(shape[1], shape[1] + nhood[e, 1]),
                        max(0, nhood[e, 2]) : min(shape[2], shape[2] + nhood[e, 2]),
                    ]
                    > 0
                )
            )

    else:
        raise RuntimeError(f"AddAffinities works only in 2 or 3 dimensions, not {dims}")

    return aff


@numba.njit(parallel=True)
def replace_values(arr, src, dst):
    shape = arr.shape
    arr = arr.ravel()
    label_map = {src[i]: dst[i] for i in range(len(src))}
    relabeled_arr = np.zeros_like(arr)

    for i in numba.prange(arr.shape[0]):
        relabeled_arr[i] = label_map.get(arr[i], arr[i])

    return relabeled_arr.reshape(shape)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Literal

from funlib.persistence.graphs import PgSQLGraphDatabase, SQLiteGraphDataBase
from funlib.persistence.graphs.graph_database import GraphDataBase
from funlib.persistence.types import Vec

from .utils import StrictBaseModel

AttrsDict = dict[str, Any]


class DB(StrictBaseModel, ABC):
    """
    A base class for defining the common attributes and methods for all
    database types.
    """

    node_attrs: dict[str, str | int] | None = None
    """
    node_attrs defines the custom node attributes you may want to save for
    your super voxels. An example of some custom node attrs is provided here:

        `node_attrs={"raw_intensity": "float"}`

    This adds a float column with name "raw_intensity" to the nodes table
    in your graph db. You can later use this attribute by name to use this
    value in a blockwise task.
    """
    edge_attrs: dict[str, str | int] | None = None
    """
    edge_attrs defines the custom edge attributes you may want to save between
    your super voxels. An example of some custom edge attrs is provided here:

        `edge_attrs={"z_aff": "float", "y_aff": "float", "x_aff": "float"}`

    This adds a float columns with names ["{zyx}_affinity"] to the edges table
    in your graph db. You can later use these attributes by name to use these
    values in a blockwise task.
    """

    ndim: int = 3
    """
    The dimensionality of your spatial data. This is used to define the number
    of coordinates used to index your node locations.
    """

    @property
    def default_node_attrs(self) -> AttrsDict:
        """
        The type definitions for the default node attributes stored in our database.

        Our DBs will always store supervoxels with the following attributes:

        position: `(float, float, float)`:

            The center of mass of the supervoxel

        size: `int`:

            the number of voxels in this supervoxel

        filtered: `bool`:

            whether or not this fragment has been filtered out

        """
        return {"position": Vec(float, self.ndim), "size": int, "filtered": bool}

    @property
    def default_edge_attrs(self) -> AttrsDict:
        """
        The type definitions for the default edge attributes stored in our database.

        Our DBs will always store edges between supervoxels with the following attributes:

        distance: `float`:

            The distance between center of masses of our super voxels

        """
        return {"distance": float}

    @property
    def graph_attrs(self) -> tuple[AttrsDict, AttrsDict]:
        """
        Get all node and edge attributes including default and user provided attributes
        """
        node_attrs = self.node_attrs if self.node_attrs is not None else {}
        parsed_node_attrs = {
            k: (Vec(float, v) if isinstance(v, int) else eval(v))
            for k, v in node_attrs.items()
        }
        parsed_node_attrs = {**self.default_node_attrs, **parsed_node_attrs}
        edge_attrs = self.edge_attrs if self.edge_attrs is not None else {}
        parsed_edge_attrs = {
            k: (Vec(float, v) if isinstance(v, int) else eval(v))
            for k, v in edge_attrs.items()
        }
        parsed_edge_attrs = {**self.default_edge_attrs, **parsed_edge_attrs}
        return parsed_node_attrs, parsed_edge_attrs

    @abstractmethod
    def open(self, mode: str = "r") -> GraphDataBase:
        """
        Return a `funlib.persistence.graphs.GraphDB` instance.
        """
        pass

    def init(self) -> None:
        """
        Create database if it doesn't exist yet.
        """
        try:
            self.open("r")
        except RuntimeError:
            self.open("w")

    @abstractmethod
    def drop(self) -> None:
        """
        Drop all nodes (supervoxels), edges, and metadata
        """
        pass

    @abstractmethod
    def drop_edges(self) -> None:
        """
        Drop all edges between nodes (supervoxels) leaving the nodes and metadata
        in tact.
        """
        pass

    @property
    @abstractmethod
    def id(self) -> str:
        """
        A unique identifier for this databse.
        """
        pass


class SQLite(DB):
    """
    An SQLite database for storing and retrieving graph data.
    """

    db_type: Literal["sqlite"] = "sqlite"
    """
    A literal used for pydantic serialization and deserialization of
    DB Union types.
    """
    path: Path
    """
    The path to the SQLite db file to use.
    """

    @property
    def id(self) -> str:
        return self.path.stem

    def open(self, mode="r") -> SQLiteGraphDataBase:
        node_attrs, edge_attrs = self.graph_attrs

        if not self.path.parent.exists():
            self.path.parent.mkdir(parents=True)

        return SQLiteGraphDataBase(
            self.path,
            position_attribute="position",
            node_attrs=node_attrs,
            edge_attrs=edge_attrs,
            mode=mode,
        )

    def drop(self) -> None:
        if self.path.exists():
            self.path.unlink()
        meta_path = self.path.parent / f"{self.id}-meta.json"
        if meta_path.exists():
            meta_path.unlink()

    def drop_edges(self) -> None:
        try:
            db = self.open("r+")
            db._drop_edges()
            db._create_tables()
        except RuntimeError:
            pass


class PostgreSQL(DB):
    """
    A PostgreSQL database for storing and retrieving graph data.
    """

    db_type: Literal["postgresql"] = "postgresql"
    host: str = "localhost"
    name: str = "volara"
    user: str | None = None
    password: str | None = None

    @property
    def id(self) -> str:
        return self.name

    def open(self, mode="r") -> PgSQLGraphDatabase:
        node_attrs, edge_attrs = self.graph_attrs

        return PgSQLGraphDatabase(
            db_host=self.host,
            db_name=self.name,
            db_user=self.user,
            db_password=self.password,
            position_attribute="position",
            node_attrs=node_attrs,
            edge_attrs=edge_attrs,
            mode=mode,
        )

    def drop(self) -> None:
        try:
            db = self.open("r+")
            db._drop_tables()
            db._create_tables()
        except RuntimeError:
            # DB doesn't exist yet
            pass

    def drop_edges(self) -> None:
        try:
            db = self.open("r+")
            db._drop_edges()
            db._create_tables()
        except RuntimeError:
            # DB doesn't exist yet
            pass



from contextlib import contextmanager
from shutil import rmtree
from typing import Literal

import numpy as np
from funlib.geometry import Coordinate, Roi

from ..datasets import Dataset, Labels, Raw
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask


class Threshold(BlockwiseTask):
    """
    Blockwise threshold an array
    """

    task_type: Literal["threshold"] = "threshold"
    in_data: Raw
    """
    The dataset to threshold.
    """
    mask: Labels
    """
    The output thresholded dataset saved as a mask.
    """
    threshold: float
    """
    The threshold value to apply to your dataset.
    """
    block_size: PydanticCoordinate

    fit: Literal["shrink"] = "shrink"
    read_write_conflict: Literal[False] = False
    _out_array_dtype: np.dtype = np.dtype(np.uint8)

    @property
    def task_name(self) -> str:
        return f"{self.mask.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.in_data.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def voxel_size(self) -> Coordinate:
        return self.in_data.array("r").voxel_size

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.voxel_size

    @property
    def context_size(self) -> Coordinate:
        return Coordinate((0,) * self.write_size.dims)

    @property
    def output_datasets(self) -> list[Dataset]:
        return [self.mask]

    def drop_artifacts(self):
        rmtree(self.mask.store, ignore_errors=True)

    def init(self):
        self.init_out_array()

    def init_out_array(self):
        in_data = self.in_data.array("r")
        self.mask.prepare(
            self.write_roi.shape / self.voxel_size,
            self.write_size / self.voxel_size,
            offset=self.write_roi.offset,
            voxel_size=self.voxel_size,
            units=in_data.units,
            axis_names=in_data.axis_names,
            types=in_data.types,
            dtype=self._out_array_dtype,
        )

    @contextmanager
    def process_block_func(self):
        frags = self.in_data.array("r")
        segs = self.mask.array("r+")

        def process_block(block):
            segs[block.write_roi] = frags[block.write_roi] > self.threshold

        yield process_block



from pathlib import Path

import daisy
import numpy as np
from funlib.geometry import Coordinate, Roi
from funlib.persistence.arrays import prepare_ds

from volara.datasets import Affs, Dataset, Labels, Raw
from volara.dbs import SQLite

BLOCK = daisy.Block(
    total_roi=Roi((0, 0), (10, 10)),
    read_roi=Roi((0, 0), (10, 10)),
    write_roi=Roi((0, 0), (10, 10)),
)


def build_zarr(
    tmpdir: Path,
    name: str,
    data: np.ndarray,
    spatial_dims: int,
    neighborhood: list[Coordinate] | None = None,
) -> Dataset:
    arr = prepare_ds(
        tmpdir / "test_data.zarr" / name,
        shape=data.shape,
        voxel_size=Coordinate((1,) * spatial_dims),
        dtype=data.dtype,
        mode="w",
    )
    arr[:] = data
    dataset_type = {
        "raw": Raw,
        "probs": Raw,
        "affs": Affs,
        "frags": Labels,
        "segments": Labels,
        "labels": Labels,
    }[name]

    kwargs = {}
    if neighborhood is not None:
        kwargs["neighborhood"] = neighborhood
    return dataset_type(store=tmpdir / "test_data.zarr" / name, **kwargs)


def build_db(tmpdir: Path) -> SQLite:
    data_dir = tmpdir / "test_data.zarr"
    if not data_dir.exists():
        data_dir.mkdir()
    db_config = SQLite(
        path=tmpdir / "test_data.zarr" / "db.sqlite",
        node_attrs={"raw_intensity": 1},
        edge_attrs={
            "y_aff": "float",
        },
        ndim=2,
    )
    db_config.init()
    return db_config


def test_distance_agglom(tmpdir):
    pass


def test_distance_agglom_simple(tmpdir):
    pass


def test_dummy(tmpdir):
    pass


def test_extract_frags(tmpdir):
    pass


def test_lut(tmpdir):
    pass


def test_predict(tmpdir):
    pass


def test_seeded_extract_frags(tmpdir):
    pass


def test_threshold(tmpdir):
    pass



import logging
from contextlib import contextmanager
from shutil import rmtree
from typing import Annotated, Literal

import daisy
import mwatershed as mws
import numpy as np
from funlib.geometry import Coordinate, Roi
from funlib.persistence import Array
from pydantic import Field
from scipy.ndimage import gaussian_filter, measurements
from skimage.measure import label as relabel
from skimage.morphology import remove_small_objects

from ..datasets import Affs, Labels, Raw
from ..dbs import PostgreSQL, SQLite
from ..tmp import replace_values
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask

logger = logging.getLogger(__file__)


class ExtractFrags(BlockwiseTask):
    """
    A task for extracting fragments from affinities.
    Internally it uses mutex watershed to agglomerate fragments.
    """

    task_type: Literal["extract-frags"] = "extract-frags"
    db: Annotated[
        PostgreSQL | SQLite,
        Field(discriminator="db_type"),
    ]
    """
    The database into which we will store node centers along with statistics
    such as fragment size, mean intensity, etc.
    """
    affs_data: Affs
    """
    The affinities dataset that we will use to extract fragments.
    """
    frags_data: Labels
    """
    The output dataset that will contain the extracted fragments.
    """
    mask_data: Raw | None = None
    """
    An optional mask that will be used to ignore some affinities.
    """
    block_size: PydanticCoordinate
    context: PydanticCoordinate
    bias: list[float]
    """
    The merge/split bias for the affinities. This should be a vector of length equal to the
    size of the neighborhood with one bias per offset. This allows you to have a merge preferring
    bias for short range affinites and a split preferring bias for long range affinities.

    Example:
    Assuming you trained affs [(0, 1), (1, 0), (0, 4), (4, 0)] for a 2D dataset, you can set the bias
    to [-0.2, -0.2, -0.8, -0.8]. This will bias you towards merging on short range affinities and splitting on
    long range affinities which has been shown to work well.
    """
    strides: list[PydanticCoordinate] | None = None
    """
    The strides to use for each affinity offset in the mutex watershed algorithm. If you have
    long range affinities it can be heplful to ignore some percentage of them to avoid excessive
    splits, so you may want to use only every other voxel in the z direction for example.

    Example:
    Assuming you trained affs [(0, 1), (1, 0), (0, 4), (4, 0)] for a 2D dataset, you can set the strides
    to [(1, 1), (1, 1), (2, 2), (2, 2)]. This will result in only 1 in every 4 long range affinities
    being used in the mutex watershed algorithm resulting in fewer splits (assuming you biased long
    range affinities towards splitting).
    """
    sigma: PydanticCoordinate | None = None
    """
    The amplitude of the smoothing kernel to apply to the affinities before watershed.
    This can help agglomerate fragments from the inside out to avoid a small merge error
    causing a large fragment to split in half.
    """
    noise_eps: float | None = None
    """
    The amplitude of the random noise to add to the affinities before watershed. This
    also helps avoid streak like fragment artifacts from processing affinities in a fifo order.
    """
    filter_fragments: float = 0.0
    """
    The minimum average affinity value for a fragment to be considered valid. If the average
    affinity value is below this threshold the fragment will be removed.
    """
    remove_debris: int = 0
    """
    The minimum size of a fragment to be considered valid. If the fragment is smaller than this
    value it will be removed.
    """
    randomized_strides: bool = False
    """
    If using strides, you may want to switch from a grided stride to a random probability of
    filtering out an affinity. This can help avoid grid artifacts in the fragments.
    """

    fit: Literal["shrink"] = "shrink"
    read_write_conflict: Literal[False] = False
    _out_array_dtype: np.dtype = np.dtype(np.uint64)

    @property
    def neighborhood(self):
        return self.affs_data.neighborhood

    @property
    def task_name(self) -> str:
        return f"{self.frags_data.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.affs_data.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.affs_data.array("r").voxel_size

    @property
    def context_size(self) -> Coordinate:
        return self.context * self.affs_data.array("r").voxel_size

    @property
    def num_voxels_in_block(self) -> int:
        return int(np.prod(self.block_size))

    @property
    def voxel_size(self) -> Coordinate:
        return self.affs_data.array("r").voxel_size

    def drop_artifacts(self):
        try:
            rmtree(self.frags_data.store)
        except FileNotFoundError:
            pass
        self.db.drop()

    def init(self):
        self.db.init()
        self.init_out_array()

    def init_out_array(self):
        in_data = self.affs_data.array("r")
        self.frags_data.prepare(
            self.write_roi.shape / self.voxel_size,
            self.write_size / self.voxel_size,
            self.write_roi.offset,
            self.voxel_size,
            units=in_data.units,
            axis_names=in_data.axis_names[1:],
            types=in_data.types[1:],
            dtype=self._out_array_dtype,
        )

    def filter_avg_fragments(self, affs, fragments_data, filter_value):
        # tmp (think about this)
        average_affs = np.mean(affs[0:3], axis=0)

        filtered_fragments = []

        fragment_ids = np.unique(fragments_data)

        for fragment, mean in zip(
            fragment_ids, measurements.mean(average_affs, fragments_data, fragment_ids)
        ):
            if mean < filter_value:
                filtered_fragments.append(fragment)

        filtered_fragments = np.array(filtered_fragments, dtype=fragments_data.dtype)
        replace = np.zeros_like(filtered_fragments)
        replace_values(fragments_data, filtered_fragments, replace)

    def get_fragments(self, affs_data):
        fragments_data = self.compute_fragments(affs_data)

        # # mask fragments if provided
        # if mask is not None:
        #     fragments_data *= mask_data.astype(np.uint64)

        # filter fragments
        if self.filter_fragments > 0:
            self.filter_avg_fragments(affs_data, fragments_data, self.filter_fragments)

        # remove small debris
        if self.remove_debris > 0:
            fragments_dtype = fragments_data.dtype
            fragments_data = fragments_data.astype(np.int64)
            fragments_data = remove_small_objects(
                fragments_data, min_size=self.remove_debris
            )
            fragments_data = fragments_data.astype(fragments_dtype)

        return fragments_data

    def compute_fragments(self, affs_data):
        if self.sigma is not None:
            # add 0 for channel dim
            sigma = (0, *self.sigma)
        else:
            sigma = None

        # add some random noise to affs (this is particularly necessary if your affs are
        #  stored as uint8 or similar)
        # If you have many affinities of the exact same value the order they are processed
        # in may be fifo, so you can get annoying streaks.

        ### tmp comment out ###

        shift = np.zeros_like(affs_data)

        if self.noise_eps is not None:
            shift += np.random.randn(*affs_data.shape) * self.noise_eps

        #######################

        # add smoothed affs, to solve a similar issue to the random noise. We want to bias
        # towards processing the central regions of objects first.

        ### tmp comment out ###

        if sigma is not None:
            shift += gaussian_filter(affs_data, sigma=sigma) - affs_data

        #######################
        shift += np.array([self.bias]).reshape(
            (-1, *((1,) * (len(affs_data.shape) - 1)))
        )

        fragments_data = mws.agglom(
            (affs_data + shift).astype(np.float64),
            offsets=self.neighborhood,
            strides=self.strides,
            randomized_strides=self.randomized_strides,
        )

        return fragments_data

    def watershed_in_block(
        self,
        block: daisy.Block,
        affs: Array,
        frags: Array,
        rag_provider,
        mask: Array | None = None,
    ):
        # todo: simplify or break into more functions

        affs_data = affs.to_ndarray(block.read_roi, fill_value=0)

        if affs.dtype == np.uint8:
            max_affinity_value = 255.0
            affs_data = affs_data.astype(np.float64)
        else:
            max_affinity_value = 1.0

        if affs_data.max() < 1e-3:
            return

        affs_data /= max_affinity_value

        if mask is not None:
            logger.debug("reading mask from %s", block.read_roi)
            mask_data = mask.to_ndarray(block.read_roi, fill_value=0)

            if len(mask_data.shape) == block.read_roi.dims + 1:
                # assume masking with raw data where data > 0
                mask_data = (np.min(mask_data, axis=0) > 0).astype(np.uint8)

            if np.max(mask_data) == 255:
                # should be ones
                mask_data = (mask_data > 0).astype(np.uint8)

            logger.debug("masking affinities")
            affs_data *= mask_data

        fragments_data = self.get_fragments(affs_data)

        fragments = Array(
            fragments_data, offset=block.read_roi.offset, voxel_size=frags.voxel_size
        )

        # crop fragments to write_roi
        fragments_data = fragments.to_ndarray(block.write_roi)
        max_id = fragments_data.max()

        fragments_data, max_id = relabel(fragments_data, return_num=True)
        assert max_id < self.num_voxels_in_block, f"max_id: {max_id}"

        # ensure unique IDs
        id_bump = block.block_id[1] * self.num_voxels_in_block
        fragments_data[fragments_data > 0] += id_bump

        # store fragments
        frags[block.write_roi] = fragments_data

        # following only makes a difference if fragments were found
        if fragments_data.max() == 0:
            return

        fragment_ids, counts = np.unique(fragments_data, return_counts=True)
        logger.info("Found %d fragments", len(fragment_ids))
        fragment_ids, counts = zip(
            *[(f, c) for f, c in zip(fragment_ids, counts) if f > 0]
        )
        centers_of_masses = measurements.center_of_mass(
            np.ones_like(fragments_data), fragments_data, fragment_ids
        )

        fragment_centers = {
            fragment_id: {
                "center": block.write_roi.get_offset()
                + affs.voxel_size * Coordinate(center),
                "size": count,
            }
            for fragment_id, center, count in zip(
                fragment_ids, centers_of_masses, counts
            )
            if fragment_id > 0
        }

        # store nodes
        rag = rag_provider[block.write_roi]

        for node, data in fragment_centers.items():
            # centers
            node_attrs = {
                "position": data["center"],
            }

            node_attrs["size"] = int(data["size"])

            rag.add_node(int(node), **node_attrs)

        rag_provider.write_graph(
            rag,
            block.write_roi,
        )

    @contextmanager
    def process_block_func(self):
        affs = self.affs_data.array("r")
        frags = self.frags_data.array("r+")
        mask = self.mask_data.array("r") if self.mask_data else None

        rag_provider = self.db.open("r+")

        def process_block(block):
            self.watershed_in_block(
                block,
                affs,
                frags,
                rag_provider,
                mask=mask,
            )

        yield process_block



from contextlib import contextmanager
from shutil import rmtree
from typing import Literal

import numpy as np
from funlib.geometry import Coordinate, Roi

from ..datasets import Dataset, Labels, Raw
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask


class Argmax(BlockwiseTask):
    """
    A blockwise task that performs an argmax operation on a given set of
    probabilities and writes the result to a semantic segmentation dataset.
    """

    task_type: Literal["argmax"] = "argmax"
    probs_data: Raw
    """
    The dataset containing raw probabilities for which you want to
    compute the argmax.
    """
    sem_data: Labels
    """
    The dataset in which we will store the final semantic labels.
    """
    combine_classes: list[list[int]] | None = None
    """
    A list of lists containing the ids to combine. All channels in `combine_classes[i]`
    will be summed into a new channel `i` before computing the argmax.
    """
    block_size: PydanticCoordinate
    """
    The block size with which to chunk our argmax task.
    """
    fit: Literal["shrink"] = "shrink"
    read_write_conflict: Literal[False] = False

    @property
    def task_name(self) -> str:
        return f"{self.sem_data.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.probs_data.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def voxel_size(self) -> Coordinate:
        return self.probs_data.array("r").voxel_size

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.voxel_size

    @property
    def context_size(self) -> Coordinate:
        return Coordinate((0,) * self.write_size.dims)

    @property
    def output_datasets(self) -> list[Dataset]:
        return [self.sem_data]

    def drop_artifacts(self):
        rmtree(self.sem_data.store)

    def init(self):
        self.init_out_array()

    def init_out_array(self):
        # get data from in_array
        shape = self.write_roi.shape // self.voxel_size
        chunk_shape = self.write_size // self.voxel_size

        in_data = self.probs_data.array("r")

        self.sem_data.prepare(
            shape,
            chunk_shape,
            self.write_roi.offset,
            self.voxel_size,
            units=in_data.units,
            axis_names=in_data.axis_names[1:],
            types=in_data.types[1:],
            dtype=self._out_array_dtype,
        )

    def argmax_block(self, block, probabilities, semantic):
        probs = probabilities.to_ndarray(block.write_roi)
        if self.combine_classes is not None:
            combined = np.zeros(
                (len(self.combine_classes),) + probs.shape[1:], dtype=probs.dtype
            )
            for i, classes in enumerate(self.combine_classes):
                combined[i] = np.sum(probs[classes], axis=0)
            probs = combined
        semantic[block.write_roi] = np.argmax(probs, 0)

    @contextmanager
    def process_block_func(self):
        probabilities = self.probs_data.array("r")
        semantic = self.sem_data.array("r+")

        def process_block(block):
            self.argmax_block(block, probabilities, semantic)

        yield process_block



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from shutil import rmtree
from typing import Literal, Sequence

import numpy as np
import zarr
from funlib.geometry import Coordinate
from funlib.persistence import Array, open_ds, prepare_ds
from funlib.persistence.arrays.datasets import ArrayNotFoundError

from .utils import PydanticCoordinate, StrictBaseModel

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)


class Dataset(StrictBaseModel, ABC):
    """
    A Dataset base class that defines the common attributes and methods
    for all dataset types.
    """

    store: str | Path

    voxel_size: PydanticCoordinate | None = None
    offset: PydanticCoordinate | None = None
    axis_names: list[str] | None = None
    units: list[str] | None = None

    @property
    def name(self) -> str:
        """
        A name for this dataset. Often it is simply the name of the
        path provided as the store. We use it to differentiate between
        multiple runs of the same blockwise task on different data.
        """
        if isinstance(self.store, Path):
            return self.store.name
        else:
            return self.store.split("/")[-1]

    def drop(self) -> None:
        """
        Delete this dataset
        """
        rmtree(self.store)

    def prepare(
        self,
        shape: Sequence[int],
        chunk_shape: Sequence[int],
        offset: Coordinate,
        voxel_size: Coordinate,
        units: list[str],
        axis_names: list[str],
        types: list[str],
        dtype,
    ) -> None:
        # prepare ds
        array = prepare_ds(
            self.store,
            shape=shape,
            offset=offset,
            voxel_size=voxel_size,
            units=units,
            axis_names=axis_names,
            types=types,
            chunk_shape=chunk_shape,
            dtype=dtype,
            mode="a",
        )
        array._source_data.attrs.update(self.attrs)

    def array(self, mode: str = "r") -> Array:
        return open_ds(self.store, mode=mode)

    @property
    @abstractmethod
    def attrs(self):
        pass


class Raw(Dataset):
    """
    Represents a dataset containing raw intensities.
    Has support for sampling specific channels, normalizing
    with provided scale and shifting, or reading in normalization
    bounds from OMERO metadata.
    """

    dataset_type: Literal["raw"] = "raw"
    channels: list[int] | None = None
    ome_norm: Path | str | None = None
    scale_shift: tuple[float, float] | None = None
    stack: Dataset | None = None

    @property
    def bounds(self) -> list[tuple[float, float]] | None:
        if self.ome_norm is not None:
            array = open_ds(self.store, mode="r")
            metadata_group = zarr.open(self.ome_norm)
            channels_meta = metadata_group.attrs["omero"]["channels"]
            bounds = [
                (channels_meta[c]["window"]["min"], channels_meta[c]["window"]["max"])
                for c in range(array.data.shape[0])
            ]
            return bounds
        else:
            return None

    @property
    def attrs(self):
        attrs = {}
        if self.channels is not None:
            attrs["channels"] = self.channels
        if self.ome_norm:
            attrs["bounds"] = self.bounds
        return attrs

    def array(self, mode="r"):
        def scale_shift(data, scale_shift):
            data = data.astype(np.float32)
            scale, shift = scale_shift
            norm = data * scale + shift
            return norm

        def ome_norm(data, bounds):
            data = data.astype(np.float32)
            c, *shape = data.shape
            shift = np.array(
                [b_min for (b_min, _) in bounds], dtype=np.float32
            ).reshape(c, *((1,) * len(shape)))
            scale = np.array(
                [b_max - b_min for b_min, b_max in bounds], dtype=np.float32
            ).reshape(c, *((1,) * len(shape)))
            return (data - shift) / scale

        def stack(data, other_data):
            return np.concatenate([data, other_data], axis=0)

        metadata = {
            "voxel_size": self.voxel_size if self.voxel_size is not None else None,
            "offset": self.offset if self.offset is not None else None,
            "axis_names": self.axis_names if self.axis_names is not None else None,
            "units": self.units if self.units is not None else None,
        }

        array = open_ds(
            self.store,
            mode=mode,
            **{k: v for k, v in metadata.items() if v is not None},
        )

        if self.ome_norm:
            array.lazy_op(lambda data: ome_norm(data, self.bounds))
        if self.scale_shift is not None:
            array.lazy_op(lambda data: scale_shift(data, self.scale_shift))
        if self.channels is not None:
            array.lazy_op(np.s_[self.channels])
        if self.stack is not None:
            array.lazy_op(lambda data: stack(data, self.stack.array("r").data))

        return array


class Affs(Dataset):
    """
    Represents a dataset containing affinities.
    Requires the inclusion of the neighborhood for these
    affinities.
    """

    dataset_type: Literal["affs"] = "affs"
    neighborhood: list[PydanticCoordinate] | None = None

    @property
    def attrs(self):
        return {"neighborhood": self.neighborhood}

    def model_post_init(self, context):
        try:
            in_array = self.array("r")
        except ArrayNotFoundError as e:
            in_array = None
            if self.neighborhood is None:
                raise ValueError(
                    "Affs(..., neighborhood=?)\n"
                    "neighborhood must be provided when referencing an array that does not yet exist\n"
                ) from e
        if in_array is not None and "neighborhood" in in_array.attrs:
            neighborhood = in_array.attrs["neighborhood"]
            if self.neighborhood is None:
                self.neighborhood = list(Coordinate(offset) for offset in neighborhood)
            else:
                assert np.isclose(neighborhood, self.neighborhood).all(), (
                    f"(Neighborhood metadata) {neighborhood} != {self.neighborhood} (given Neighborhood)"
                )
        else:
            if self.neighborhood is None:
                raise ValueError(
                    "Affs(..., neighborhood=?)\n"
                    "neighborhood must be provided when referencing an affs array that does not have "
                    "a neighborhood key in the `.zattrs`"
                )
        return super().model_post_init(context)


class LSD(Dataset):
    """
    Represents a dataset containing local shape descriptors.
    """

    dataset_type: Literal["lsd"] = "lsd"

    @property
    def attrs(self):
        return {"lsds": True}


class Labels(Dataset):
    """
    Represents an integer label dataset.
    """

    dataset_type: Literal["labels"] = "labels"

    @property
    def attrs(self):
        return {}



import pytest

from volara.dbs import DB, PostgreSQL, SQLite


def psql_is_available():
    try:
        import psycopg2

        conn = psycopg2.connect(
            dbname="pytest",
        )
        conn.close()
        return True
    except psycopg2.OperationalError:
        return False
    except ImportError:
        return False


@pytest.mark.parametrize(
    "db_type",
    [
        "sqlite",
        pytest.param(
            "postgresql",
            marks=pytest.mark.skipif(
                not psql_is_available(),
                reason="PostgreSQL is not available",
            ),
        ),
    ],
)
def test_dbs(db_type: str, tmp_path):
    db: DB
    if db_type == "sqlite":
        db = SQLite(
            node_attrs={"color": 3},
            edge_attrs={
                "y_aff": "float",
            },
            ndim=2,
            path=tmp_path / "db.sqlite",
        )
    else:
        db = PostgreSQL(
            node_attrs={"color": 3},
            edge_attrs={
                "y_aff": "float",
            },
            ndim=2,
        )

    with pytest.raises(RuntimeError):
        db.open("r")

    db.init()
    graph_provider = db.open("r+")

    g = graph_provider.read_graph()
    assert g.number_of_nodes() == 0
    assert g.number_of_edges() == 0

    g.add_node(
        0,
        position=(0, 0),
        color=[155, 100, 0],
    )
    g.add_node(
        1,
        position=(1, 1),
        color=[55, 100, 155],
    )
    g.add_edge(0, 1, y_aff=0.5)
    graph_provider.write_graph(g)

    graph_provider = db.open("r")
    g2 = graph_provider.read_graph()
    assert g2.number_of_nodes() == 2
    assert g2.number_of_edges() == 1

    assert g2.nodes[0]["color"] == (155, 100, 0)
    assert g2.nodes[1]["color"] == (55, 100, 155)
    assert g2.edges[0, 1]["y_aff"] == 0.5
    assert g2.nodes[0]["position"] == (0, 0)
    assert g2.nodes[1]["position"] == (1, 1)

    db.drop_edges()
    graph_provider = db.open("r")
    g3 = graph_provider.read_graph()

    assert g3.number_of_nodes() == 2
    assert g3.number_of_edges() == 0

    db.drop()
    with pytest.raises(RuntimeError):
        db.open("r")

    graph_provider = db.open("w")



from pathlib import Path

import daisy

# default log dir
LOG_BASEDIR = Path("./volara_logs")
daisy.logging.set_log_basedir(LOG_BASEDIR)


def set_log_basedir(path: Path):
    """Set the base directory for logging (indivudal worker logs and detailed
    task summaries). If set to ``None``, all logging will be shown on the
    command line (which can get very messy).

    Default is ``./volara_logs``.
    """

    global LOG_BASEDIR

    if path is not None:
        LOG_BASEDIR = Path(path)
    else:
        raise NotImplementedError("None is not a valid log directory")
        LOG_BASEDIR = None

    daisy.logging.set_log_basedir(LOG_BASEDIR)


def get_log_basedir():
    """Get the base directory for logging (indivudal worker logs and detailed
    task summaries).

    Default is ``./volara_logs``.
    """
    global LOG_BASEDIR
    return LOG_BASEDIR



import glob
import logging
import multiprocessing
import subprocess
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING

import daisy
import numpy as np
from daisy.block import BlockStatus
from funlib.geometry import Coordinate, Roi
from funlib.math import cantor_number
from funlib.persistence import open_ds, prepare_ds

from volara.logging import get_log_basedir, set_log_basedir

from ..utils import PydanticCoordinate, StrictBaseModel
from ..workers import Worker

if TYPE_CHECKING:
    from .pipeline import Pipeline

logger = logging.getLogger(__name__)


class BlockwiseTask(StrictBaseModel, ABC):
    roi: tuple[PydanticCoordinate, PydanticCoordinate] | None = None
    """
    An optional `roi` defining the total region to output.
    """
    num_workers: int = 1
    """
    The number of workers that will be started to process blocks in parallel
    """
    num_cache_workers: int | None = None
    """
    The number of threads running the process_block function per worker.
    This allows you to start e.g. 4 gpu workers, each with 1 copy of the gpu loaded
    and running 8 threads to read, pre/postprocess, and write your data; maximizing
    your gpu utilization.
    """
    worker_config: Worker | None = None
    """
    The configuration for each worker you start. This allows you to specify
    arguments for running workers on various platforms such as slurm/lsf clusters
    or AWS EC2.
    """
    _out_array_dtype: np.dtype = np.dtype(np.uint8)
    """
    The output array data type
    """

    fit: str
    """
    The strategy to use for blocks that overhang your total write roi.
    """
    read_write_conflict: bool
    """
    Whether blocks have read/write dependencies on neighborhing blocks requiring
    a specific ordering to the block processing to compute a seamless result.
    """

    def __hash__(self):
        return hash(self.task_name)

    @property
    @abstractmethod
    def task_name(self) -> str:
        """
        A unique identifier for a task. This allows us to store log files
        in an unambiguous location as well as storing a block_done dataset
        that allows us to cache completed blocks and resume processing
        at a later time.
        """
        pass

    @property
    @abstractmethod
    def write_roi(self) -> Roi:
        """
        The total roi of any data output by a task.
        """
        pass

    @property
    @abstractmethod
    def write_size(self) -> Coordinate:
        """
        The write size of each block processed as part of a task.
        """
        pass

    @property
    @abstractmethod
    def context_size(self) -> Coordinate | tuple[Coordinate, Coordinate]:
        """
        The amount of context needed to process each block for a task.
        """
        pass

    def init(self):
        """
        Any one time initializations that need to be made before starting a
        task such as creating dbs and zarrs.
        """
        pass

    @abstractmethod
    def process_block_func(self):
        """
        A constructor for a function that will take a single block
        as input and process it.
        """
        pass

    @abstractmethod
    def drop_artifacts(self):
        """
        A helper function to reset anything produced by a task
        to a clean state equivalent to not having run the task at all
        """
        pass

    @property
    def block_write_roi(self) -> Roi:
        """
        The write roi of a block with zero offset
        """
        return Roi((0,) * self.write_size.dims, self.write_size)

    @property
    def meta_dir(self) -> Path:
        """
        The path to the meta directory where we will store log files
        and a block done cache for resuming work if processing is
        interrupted.
        """
        return get_log_basedir() / f"{self.task_name}-meta"

    @property
    def config_file(self) -> Path:
        """
        The config file that will be used to serialize this task for
        logging purposes.
        """
        return self.meta_dir / "config.json"

    @property
    def block_ds(self) -> Path:
        """
        The dataset that will be used to track which blocks have been
        successfully completed.
        """
        return self.meta_dir / "blocks_done.zarr"

    def process_roi(self, roi: Roi, context: Coordinate | None = None):
        """
        A helper function to process a given roi without needing to start a
        whole blockwise job.
        """
        block = daisy.Block(
            roi, roi if context is None else roi.grow(context, context), roi
        )
        process_block = self.process_block_func()
        process_block(block)

    def drop(self, drop_outputs: bool = False) -> None:
        """
        A helper function to drop any artifacts produced by a task
        and return to a state identical to before having executed the
        task.
        """
        # reset the blocks_done ds so that the task is rerun
        if self.meta_dir.exists():
            rmtree(self.meta_dir)
        self.drop_artifacts()

    def check_block_func(self):
        """
        A function to check whether a block has been completed.
        """

        def check_block(block):
            block_array = open_ds(self.block_ds, mode="r")
            coordinate = (
                block.write_roi.offset - block_array.offset
            ) / block_array.voxel_size
            chunk_size = Coordinate(block_array._source_data.chunks[-coordinate.dims :])
            chunk_index = coordinate // chunk_size
            chunk_key = "/".join(str(i) for i in chunk_index)
            prefix = "/".join("*" for _ in range(block_array.dims - len(chunk_index)))

            glob_pattern = f"{prefix}/{chunk_key}" if len(prefix) > 0 else chunk_key
            glob_pattern = f"{block_array._source_data.store.path}/{glob_pattern}"
            matches = glob.glob(glob_pattern, recursive=True)
            return len(list(matches)) > 0

        return check_block

    def mark_block_done_func(self):
        """
        A helper function to mark a block as completed so that it
        can be skipped if we have to pause and resume processing later.
        """

        def write_check_block(block):
            if not block.status == BlockStatus.FAILED:
                # Unless the block is explicitly marked as failed, we assume
                # successful processing if there was no error
                block_array = open_ds(self.block_ds, mode="a")
                write_roi = block.write_roi.intersect(block_array.roi)
                block_array[write_roi] = np.full(
                    write_roi.shape // block_array.voxel_size,
                    fill_value=block.block_id[1] + 1,
                )
                block.status = BlockStatus.SUCCESS

        return write_check_block

    def worker_func(self):
        """
        The function defining how workers are started.
        """
        if self.worker_config is not None:
            config_file = self.config_file

            with open(config_file, "w") as f:
                f.write(self.model_dump_json())

            logging.info("Running block with config %s..." % config_file)

            def run_worker():
                cmd = self.worker_config.get_command(config_file, self.task_name)
                return subprocess.run(cmd)

            return run_worker

        else:
            return self.process_blocks

    def process_blocks(self):
        """
        Start our workers and run through every block until a task
        is complete.
        """
        with self.process_block_func() as process_block:

            def worker_loop():
                client = daisy.Client()
                # TODO: this shouldn't be necessary, daisy should be doing this for us
                try:
                    set_log_basedir(client.context["logdir"])
                except KeyError as e:
                    raise ValueError(client.context) from e
                mark_block_done = self.mark_block_done_func()

                while True:
                    logger.info("getting block")
                    with client.acquire_block() as block:
                        logger.info(f"got block {block}")

                        if block is None:
                            break

                        process_block(block)
                        mark_block_done(block)

            if self.num_cache_workers is not None:
                workers = [
                    multiprocessing.Process(target=worker_loop)
                    for _ in range(self.num_cache_workers)
                ]

                for worker in workers:
                    worker.start()

                for worker in workers:
                    worker.join()

            else:
                worker_loop()

    def init_block_array(self):
        """
        Build the block done zarr for tracking completed blocks.
        """
        # prepare blocks done ds

        def cmin(a, b):
            return Coordinate([min(ai, bi) for ai, bi in zip(a, b)])

        def cmax(a, b):
            return Coordinate([max(ai, bi) for ai, bi in zip(a, b)])

        def gcd(a, b):
            while b:
                a, b = b, a % b
            return a

        def cgcd(a, *bs):
            while len(bs) > 0:
                b = bs[0]
                bs = bs[1:]
                a, b = cmax(a, b), cmin(a, b)
                a = Coordinate([gcd(ai, bi) for ai, bi in zip(a, b)])
            return abs(a)

        def get_dtype(write_roi, write_size):
            # need to factor in block offset, so use cantor number of last block
            # + 1 to be safe
            num_blocks = cantor_number(write_roi.shape / write_size + 1)

            for dtype in [np.uint8, np.uint16, np.uint32, np.uint64]:
                if num_blocks <= np.iinfo(dtype).max:
                    return dtype
            raise ValueError(
                f"Number of blocks ({num_blocks}) is too large for available data types."
            )

        block_voxel_size = cgcd(
            self.write_roi.offset, self.write_size, self.write_roi.shape
        )

        try:
            prepare_ds(
                self.block_ds,
                shape=self.write_roi.shape / block_voxel_size,
                offset=self.write_roi.offset,
                voxel_size=block_voxel_size,
                chunk_shape=self.write_size / block_voxel_size,
                dtype=get_dtype(self.write_roi, self.write_size),
                mode="a",
            )
        except PermissionError as e:
            # The dataset already exists but with different parameters.
            existing_block_ds = open_ds(self.block_ds, mode="r")
            error_msg = (
                f"Trying to overwrite existing {self.block_ds} array with incompatible data:\n"
                f"Shape (existing): {existing_block_ds.shape} vs (new) {self.write_roi.shape / block_voxel_size}\n"
                f"Chunk Shape (existing): {existing_block_ds._source_data.chunks} vs (new) {self.write_size / block_voxel_size}\n"
                f"Data Type (existing): {existing_block_ds.dtype} vs (new) {get_dtype(self.write_roi, self.write_size)}\n"
            )
            raise ValueError(error_msg) from e

    @contextmanager
    def task(
        self,
        upstream_tasks: daisy.Task | list[daisy.Task] | None = None,
        multiprocessing: bool = True,
    ) -> daisy.Task:
        """
        Builds a `daisy.Task` that puts together everything necessary to run a task
        blockwise.
        """

        # initialize the arrays the task operates on
        self.init_block_array()
        self.init()

        # create task
        context = self.context_size
        if not isinstance(context, Coordinate):
            assert isinstance(context, tuple)
            context_low, context_high = context[0], context[1]
        else:
            context_low, context_high = context, context

        if multiprocessing:
            process_func = self.worker_func()
        else:
            process_block_func = self.process_block_func()
            process_block = process_block_func.__enter__()
            mark_block = self.mark_block_done_func()

            def process_func(block):
                process_block(block)
                mark_block(block)

        task = daisy.Task(
            self.task_name,
            total_roi=self.write_roi.grow(context_low, context_high),
            read_roi=self.block_write_roi.grow(context_low, context_high),
            write_roi=self.block_write_roi,
            process_function=process_func,
            read_write_conflict=self.read_write_conflict,
            fit=self.fit,
            num_workers=self.num_workers,
            check_function=self.check_block_func(),
            max_retries=2,
            timeout=None,
            upstream_tasks=(
                (
                    upstream_tasks
                    if isinstance(upstream_tasks, list)
                    else [upstream_tasks]
                )
                if upstream_tasks is not None
                else None
            ),
        )

        yield task

        if not multiprocessing:
            process_block_func.__exit__(None, None, None)

    def run_blockwise(
        self,
        upstream_tasks: list[daisy.Task] | None = None,
        multiprocessing: bool = True,
    ):
        """
        Execute this task blockwise.
        """
        with self.task(upstream_tasks, multiprocessing) as task:
            if upstream_tasks is None:
                tasks = [task]
            elif isinstance(upstream_tasks, list):
                tasks = upstream_tasks + [task]
            elif isinstance(upstream_tasks, daisy.Task):
                tasks = [upstream_tasks, task]
            else:
                raise NotImplementedError(
                    f"upstream tasks {upstream_tasks} with type {type(upstream_tasks)} not supported. "
                    "Please provide a daisy.Task or a list of daisy.Task objects."
                )
            if multiprocessing:
                result = daisy.run_blockwise(tasks)  # noqa
            else:
                server = daisy.SerialServer()
                _cl_monitor = daisy.cl_monitor.CLMonitor(server)  # noqa
                result = server.run_blockwise(tasks)
            return result

    def __add__(self, other: "BlockwiseTask | Pipeline") -> "Pipeline":
        """
        The task or pipeline (`task`) gets run in series after `self`.

        This means that every node in `self` without outgoing edges
        gets an edge to all nodes in `task` without incoming edges.
        """
        from .pipeline import Pipeline

        print("Task add")

        if isinstance(other, Pipeline):
            return Pipeline(self) + other
        elif isinstance(other, BlockwiseTask):
            return Pipeline(self) + Pipeline(other)
        else:
            raise NotImplementedError(
                f"We do not support other with type {type(other)}"
            )

    def __or__(self, other: "BlockwiseTask | Pipeline") -> "Pipeline":
        """
        The task or pipeline (`task`) gets run in parallel with `self`.

        Task graphs are merged, but no edges are added.
        """
        from .pipeline import Pipeline

        print("Task or")

        if isinstance(other, Pipeline):
            return Pipeline(self) | other
        elif isinstance(other, BlockwiseTask):
            return Pipeline(self) | Pipeline(other)
        else:
            raise NotImplementedError(
                f"We do not support other with type {type(other)}"
            )



from contextlib import contextmanager
from shutil import rmtree
from typing import Literal

import numpy as np
from daisy import Block
from funlib.geometry import Coordinate, Roi
from funlib.persistence import Array
from scipy.ndimage import map_coordinates
from skimage import registration

from volara.blockwise import BlockwiseTask
from volara.datasets import Dataset, Labels, Raw
from volara.utils import PydanticCoordinate


class ComputeShift(BlockwiseTask):
    task_type: Literal["compute_shift"] = "compute_shift"
    intensities: Raw
    shifts: Raw
    block_size: PydanticCoordinate
    context: PydanticCoordinate
    mask: Labels | None = None
    overlap_ratio: float | None = None
    target: Raw | None = None
    fit: Literal["overhang"] = "overhang"
    read_write_conflict: Literal[False] = False

    @property
    def task_name(self) -> str:
        return f"{self.shifts.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.intensities.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def voxel_size(self) -> Coordinate:
        return self.intensities.array("r").voxel_size

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.voxel_size

    @property
    def context_size(self):
        return self.context * self.voxel_size

    @property
    def output_datasets(self) -> list[Dataset]:
        return [self.shifts]

    def drop_artifacts(self):
        rmtree(self.shifts.store)

    def init(self):
        self.init_out_array()

    def init_out_array(self):
        in_data = self.intensities.array("r")
        # TODO: avoid hardcoding channels first :/
        # TODO: avoid assuming channel dim exists in intensities :/
        units = in_data.units
        axis_names = in_data.axis_names[-in_data.voxel_size.dims :]
        self.shifts.prepare(
            shape=(
                self.intensities.array().shape[0],
                len(self.voxel_size),
                *self.write_roi.shape.ceil_division(self.write_size),
            ),
            chunk_shape=(
                self.intensities.array().shape[0],
                len(self.voxel_size),
                *(1,) * len(self.voxel_size),
            ),
            offset=self.write_roi.offset,
            voxel_size=self.write_size,
            units=units,
            axis_names=[in_data.axis_names[0], "axis", *axis_names],
            types=[in_data.types[0], "axis", *in_data.types[1:]],
            dtype=np.float32,
        )

    @staticmethod
    def compute_shift(
        array: np.ndarray,
        target_data: np.ndarray,
        voxel_size: np.ndarray,
        mask: np.ndarray | None = None,
        overlap_ratio: float | None = None,
    ) -> np.ndarray:
        C, Z, Y, X = array.shape
        assert target_data.shape == (1, Z, Y, X) or target_data.shape == (Z, Y, X)
        if target_data.shape == (1, Z, Y, X):
            target_data = target_data[0]

        shift_data = np.zeros((C, 3, 1, 1, 1))
        if mask is not None and mask.max() == 0:
            return shift_data

        for c in range(0, C):
            # compute shift between fixed image and moving image
            shift_xyz, error, diffphase = registration.phase_cross_correlation(
                reference_image=target_data,
                moving_image=array[c],
                upsample_factor=3,
                moving_mask=mask,
                reference_mask=mask,
                overlap_ratio=overlap_ratio,
            )
            shift_data[c, :, 0, 0, 0] = shift_xyz * voxel_size
        return shift_data

    @contextmanager
    def process_block_func(self):
        # TODO: read from in_array_config
        in_array = self.intensities.array("r")
        out_array = self.shifts.array("a")

        if self.target is not None:
            target_array = self.target.array("r")

        if self.mask is not None:
            mask_array = self.mask.array("r")

        def process_block(block: Block):
            valid_read_roi = block.read_roi.intersect(in_array.roi)
            in_data = in_array.to_ndarray(roi=valid_read_roi, fill_value=0)
            if self.target is not None:
                target_data = target_array.to_ndarray(roi=valid_read_roi, fill_value=0)
            else:
                target_data = in_data[0]
            if self.mask is not None:
                mask_data = mask_array.to_ndarray(roi=valid_read_roi, fill_value=0)
            else:
                mask_data = None
            shifts = self.compute_shift(
                in_data,
                target_data,
                np.array(self.voxel_size),
                mask_data,
                self.overlap_ratio,
            )
            shift_array = Array(
                shifts,
                offset=block.write_roi.offset,
                voxel_size=block.write_roi.shape,
            )
            write_data = shift_array.to_ndarray(block.write_roi)
            out_array[block.write_roi] = write_data

        yield process_block


class ApplyShift(BlockwiseTask):
    task_type: Literal["apply_shifts"] = "apply_shifts"
    intensities: Raw
    shifts: Raw
    aligned: Raw
    fit: Literal["overhang"] = "overhang"
    read_write_conflict: Literal[False] = False
    interp_shifts: Raw | None = None
    shift_threshold: float | None = None

    @property
    def task_name(self) -> str:
        return f"{self.aligned.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.intensities.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def voxel_size(self) -> Coordinate:
        return self.intensities.array("r").voxel_size

    @property
    def context(self):
        return self.shifts.array("r").voxel_size // self.voxel_size

    @property
    def block_size(self) -> Coordinate:
        return self.shifts.array("r").voxel_size // self.voxel_size

    @property
    def context_size(self):
        return self.context * self.voxel_size

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.voxel_size

    @property
    def output_datasets(self) -> list[Dataset]:
        return [self.shifts]

    def drop_artifacts(self):
        rmtree(self.aligned.store)
        if self.interp_shifts is not None:
            rmtree(self.interp_shifts.store)

    def init(self):
        self.init_out_array()

    def init_out_array(self):
        in_array = self.intensities.array()

        voxel_shape = self.write_roi.shape / self.voxel_size

        self.aligned.prepare(
            shape=(in_array.shape[0], *voxel_shape),
            chunk_shape=(
                in_array.shape[0],
                *self.block_size,
            ),
            offset=self.write_roi.offset,
            voxel_size=self.voxel_size,
            units=self.intensities.units,
            axis_names=in_array.axis_names,
            dtype=np.uint8,
        )

        # TODO: avoid hardcoding channels first and assuming channel dim exists in intensities :/

        if self.interp_shifts is not None:
            self.interp_shifts.prepare(
                shape=(in_array.shape[0], in_array.voxel_size.dims, *voxel_shape),
                chunk_shape=(
                    in_array.shape[0],
                    in_array.voxel_size.dims,
                    *self.block_size,
                ),
                offset=self.write_roi.offset,
                voxel_size=self.voxel_size,
                units=self.intensities.units,
                axis_names=[in_array.axis_names[0], "axis", *in_array.axis_names[1:]],
                dtype=np.float32,
            )

    @staticmethod
    def apply_shift(
        intensities: np.ndarray,
        shifts: np.ndarray,
        voxel_write_roi: Roi,
        voxel_size: np.ndarray,
        shift_threshold: float | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        C, Z, Y, X = intensities.shape
        DZYX = 3  # shift in Z, Y, X
        BZ, BY, BX = 3, 3, 3
        assert shifts.shape == (C, DZYX, BZ, BY, BX)

        aligned = np.zeros((C, *voxel_write_roi.shape), dtype=intensities.dtype)
        interpolated_shifts = np.zeros(
            (C, DZYX, *voxel_write_roi.shape), dtype=np.float32
        )

        for c in range(0, C):
            channel_shifts = shifts[c]
            if shift_threshold is not None:
                max_shift = np.repeat(
                    np.abs(channel_shifts).max(axis=0, keepdims=True), 3, axis=0
                )
                mask = max_shift > shift_threshold
                channel_shifts[mask] = 0

            coordinates = np.meshgrid(
                np.linspace(0.0, 2.0, 3),
                *[np.linspace(2 / 3, 4 / 3, axis_len // 3) for axis_len in [Z, Y, X]],
                indexing="ij",
            )
            stacked_coordinates = np.stack(coordinates)

            # Interpolate the distances to the original pixel coordinates
            interp_shifts = map_coordinates(
                channel_shifts,
                coordinates=stacked_coordinates,
                order=1,
            )

            interpolated_shifts[c] = interp_shifts

            coordinates = np.meshgrid(
                *[
                    np.linspace(axis_len // 3, (2 * axis_len) // 3 - 1, axis_len // 3)
                    for axis_len in [Z, Y, X]
                ],
                indexing="ij",
            )
            stacked_coordinates = np.stack(coordinates)

            interpolated_coords = stacked_coordinates - (
                interp_shifts / voxel_size.reshape(-1, *((1,) * len(voxel_size)))
            )
            aligned_intensities = map_coordinates(
                intensities[c], interpolated_coords, order=1
            )
            aligned[c] = aligned_intensities
        return aligned, interpolated_shifts

    @contextmanager
    def process_block_func(self):
        # TODO: read from in_array_config
        in_array = self.intensities.array("r")
        shift_array = self.shifts.array("r")
        out_array = self.aligned.array("a")
        if self.interp_shifts is not None:
            out_interp_shifts = self.interp_shifts.array("a")

        def process_block(block: Block):
            in_shift = shift_array.to_ndarray(roi=block.read_roi, fill_value=0)

            in_data = in_array.to_ndarray(roi=block.read_roi, fill_value=0)
            aligned, interp_shifts = self.apply_shift(
                in_data,
                in_shift,
                block.write_roi / self.voxel_size,
                np.array(self.voxel_size),
                self.shift_threshold,
            )
            aligned = np.clip(
                aligned * 255,
                0,
                255,
            ).astype(np.uint8)
            aligned_array = Array(
                aligned,
                offset=block.write_roi.offset,
                voxel_size=in_array.voxel_size,
            )
            write_roi = block.write_roi.intersect(out_array.roi)
            write_data = aligned_array.to_ndarray(write_roi)
            out_array[write_roi] = write_data

            if self.interp_shifts is not None:
                interp_shifts_array = Array(
                    interp_shifts,
                    offset=block.write_roi.offset,
                    voxel_size=in_array.voxel_size,
                )
                write_data = interp_shifts_array.to_ndarray(write_roi)
                out_interp_shifts[write_roi] = write_data

        yield process_block



import logging
from pathlib import Path

import click


@click.group()
@click.option(
    "--log-level",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    default="INFO",
)
def cli(log_level: str) -> None:
    logging.basicConfig(level=getattr(logging, log_level.upper()))


@cli.command()
@click.option(
    "-c", "--config-file", required=True, type=click.Path(exists=True, dir_okay=False)
)
def blockwise_worker(config_file: Path) -> None:
    import json
    from pathlib import Path

    from volara.blockwise import BlockwiseTask, get_blockwise_tasks_type

    config_file = Path(config_file)
    config_json = json.loads(config_file.open("r").read())

    BlockwiseTasks = get_blockwise_tasks_type()
    config = BlockwiseTasks.validate_python(config_json)
    assert isinstance(config, BlockwiseTask)
    config.process_blocks()



from pathlib import Path

import numpy as np
import pytest
from funlib.geometry import Coordinate
from funlib.persistence import Array, prepare_ds
from scipy.ndimage import label
from skimage import data
from skimage.filters import gaussian

from volara.blockwise import AffAgglom
from volara.datasets import Affs, Labels
from volara.dbs import SQLite
from volara.tmp import seg_to_affgraph


@pytest.fixture()
def cell(tmp_path) -> tuple[Array, Path]:
    cell_data = (data.cells3d().transpose((1, 0, 2, 3)) / 256).astype(np.uint8)
    cell_path = tmp_path / "cells3d.zarr/raw"

    # Handle metadata
    # This is a single sample image, so we don't need an
    # offset to line it up with other samples.
    offset = Coordinate(0, 0, 0)
    # voxel size in nanometers
    voxel_size = Coordinate(290, 260, 260)
    # By convention we add a '^' to non spatial axes
    axis_names = ["c^", "z", "y", "x"]
    # units for each spatial axis
    units = ["nm", "nm", "nm"]

    # Creates the zarr array with appropriate metadata
    cell_array = prepare_ds(
        cell_path,
        cell_data.shape,
        offset=offset,
        voxel_size=voxel_size,
        axis_names=axis_names,
        units=units,
        mode="w",
        dtype=np.uint8,
    )

    # Saves the cell data to the zarr array
    cell_array[:] = cell_data
    return cell_array, cell_path


@pytest.fixture()
def mask(tmp_path, cell: tuple[Array, Path]) -> tuple[Array, Path]:
    cell_array, _cell_path = cell
    mask_path = tmp_path / "cells3d.zarr/mask"
    # generate and save some psuedo gt data
    mask_array = prepare_ds(
        mask_path,
        cell_array.shape[1:],
        offset=cell_array.offset,
        voxel_size=cell_array.voxel_size,
        axis_names=cell_array.axis_names[1:],
        units=cell_array.units,
        mode="w",
        dtype=np.uint8,
    )
    cell_mask = np.clip(gaussian(cell_array[1] / 255.0, sigma=1), 0, 255) * 255 > 30
    not_membrane_mask = (
        np.clip(gaussian(cell_array[0] / 255.0, sigma=1), 0, 255) * 255 < 10
    )
    mask_array[:] = cell_mask * not_membrane_mask
    return mask_array, mask_path


@pytest.fixture()
def labels(tmp_path, mask: tuple[Array, Path]) -> tuple[Array, Path]:
    mask_array, _mask_array_path = mask
    labels_path = tmp_path / "cells3d.zarr/labels"
    # generate labels via connected components
    # generate and save some psuedo gt data
    labels_array = prepare_ds(
        labels_path,
        mask_array.shape,
        offset=mask_array.offset,
        voxel_size=mask_array.voxel_size,
        axis_names=mask_array.axis_names,
        units=mask_array.units,
        mode="w",
        dtype=np.uint8,
    )
    labels_array[:] = label(mask_array[:])[0]
    return labels_array, labels_path


@pytest.fixture()
def affs(tmp_path, labels: tuple[Array, Path]) -> tuple[Array, Path]:
    labels_array, _ = labels
    affs_path = tmp_path / "cells3d.zarr/affs"
    # generate affinity graph
    affs_array = prepare_ds(
        affs_path,
        (3,) + labels_array.shape,
        offset=labels_array.offset,
        voxel_size=labels_array.voxel_size,
        axis_names=["neighborhood^"] + labels_array.axis_names,
        units=labels_array.units,
        mode="w",
        dtype=np.uint8,
    )
    affs_array[:] = (
        seg_to_affgraph(labels_array[:], nhood=[[1, 0, 0], [0, 1, 0], [0, 0, 1]]) * 255
    )
    return affs_array, affs_path


@pytest.mark.skip(reason="pytest quitting for some reason")
def test_aff_agglom(affs, labels, tmp_path):
    affs_array, affs_path = affs
    labels_array, labels_path = labels
    db = SQLite(
        path=tmp_path / "db.sqlite",
        node_attrs={"xy_aff": "float", "z_aff": "float"},
    )
    affs_config = AffAgglom(
        db=db,
        affs_data=Affs(store=affs_path, neighborhood=[[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        frags_data=Labels(store=labels_path),
        block_size=(20, 20, 20),
        context=(2, 2, 2),
        scores={"xy_aff": [(1, 0, 0), (0, 1, 0)], "z_aff": [(0, 0, 1)]},
    )
    affs_config.run_blockwise(multiprocessing=False)

    g = db.open("r").read_graph()
    assert g.number_of_edges() == 0


def test_argmax():
    pass


def test_distance_agglom():
    pass


def test_extract_frags():
    pass


def test_global_mws():
    pass


def test_lut():
    pass


def test_predict():
    pass


def test_psuedo_affs():
    pass


def test_seeded_extract_frags():
    pass



import logging
from contextlib import contextmanager
from shutil import rmtree
from typing import Annotated, Literal

import daisy
import mwatershed as mws
import numpy as np
from funlib.geometry import Coordinate, Roi
from funlib.persistence import Array
from pydantic import Field
from scipy.ndimage import gaussian_filter

from ..datasets import Affs, Dataset, Labels
from ..dbs import PostgreSQL, SQLite
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask

logger = logging.getLogger(__file__)


class SeededExtractFrags(BlockwiseTask):
    """
    Extract fragments from affinities using a set of skeletons as a supervising signal.
    Any voxel that intersects with a node placed on a skeleton is guaranteed to be assigned
    the label of the skeleton it intersects with. The affinities are used to fill out
    the rest of the segment to get a full volume representation of your skeletons.

    Any fragment that does not intersect with a skeleton is discarded.
    """

    task_type: Literal["seeded-extract-frags"] = "seeded-extract-frags"
    affs_data: Affs
    """
    The affinities dataset that will be used to expand the skeletons to full segments.
    """
    segs_data: Labels
    """
    The segmentations dataset that will contain the final segmentations.
    """
    mask_data: Labels | None = None
    """
    An optional mask for masking out the affinities.
    """
    block_size: PydanticCoordinate
    context: PydanticCoordinate
    bias: list[float]
    """
    The bias terms to be used for each offset in the affinities neighborhoood.
    """
    strides: list[PydanticCoordinate] | None = None
    """
    The strides with which to filter each offset in the affinities neighborhood.
    """
    graph_db: Annotated[
        PostgreSQL | SQLite,
        Field(discriminator="db_type"),
    ]
    """
    The graph database containing the skeletons to use as a supervising signale.
    """
    randomized_strides: bool = False
    """
    Whether or not to convert the strides from a grid like filter to a random probability
    of filtering out each affinity edge.
    """

    fit: Literal["shrink"] = "shrink"
    read_write_conflict: Literal[False] = False
    _out_array_dtype: np.dtype = np.dtype(np.uint64)

    @property
    def task_name(self) -> str:
        return f"{self.segs_data.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        if self.roi is not None:
            return self.affs_data.array("r").roi.intersect(
                Roi(self.roi[0], self.roi[1])
            )
        else:
            return self.affs_data.array("r").roi

    @property
    def voxel_size(self) -> Coordinate:
        return self.affs_data.array("r").voxel_size

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.affs_data.array("r").voxel_size

    @property
    def context_size(self) -> Coordinate:
        return self.context * self.voxel_size

    @property
    def output_datasets(self) -> list[Dataset]:
        return [self.segs_data]

    def drop_artifacts(self):
        rmtree(self.segs_data.store)

    def init(self):
        self.init_out_array()

    def init_out_array(self):
        # get data from in_array
        in_data = self.affs_data.array("r")
        voxel_size = in_data.voxel_size

        self.segs_data.prepare(
            self.write_roi.shape / voxel_size,
            self.block_write_roi.shape / voxel_size,
            offset=self.write_roi.offset,
            voxel_size=voxel_size,
            units=in_data.units,
            axis_names=in_data.axis_names[1:],
            types=in_data.types[1:],
            dtype=self._out_array_dtype,
        )

    @contextmanager
    def process_block_func(self):
        affs_array = self.affs_data.array("r")
        segs_array = self.segs_data.array("r+")
        mask_array = self.mask_data.array("r") if self.mask_data is not None else None

        graph_provider = self.graph_db.open("r")

        def process_block(block: daisy.Block):
            affs = affs_array.to_ndarray(block.read_roi, fill_value=0)
            if mask_array is not None:
                affs *= mask_array.to_ndarray(block.read_roi, fill_value=0) > 0
            if affs.max() == 0:
                return
            graph = graph_provider.read_graph(block.read_roi)
            seeds = np.zeros(affs.shape[1:], dtype=np.uint64)
            unique_seeds = set()
            for _, node_attrs in graph.nodes(data=True):
                if "position" not in node_attrs:
                    continue
                pos = Coordinate(node_attrs["position"])
                if block.read_roi.contains(pos):
                    pos -= block.read_roi.offset
                    pos /= affs_array.voxel_size
                    seeds[tuple(pos)] = int(node_attrs["skeleton_id"])
                    unique_seeds.add(int(node_attrs["skeleton_id"]))

            if len(unique_seeds) == 0:
                return

            if affs.dtype == np.uint8:
                max_affinity_value = 255.0
                affs = affs.astype(np.float64)
            else:
                max_affinity_value = 1.0

            if affs.max() < 1e-3:
                return

            affs /= max_affinity_value

            sigma = (0, 6, 9, 9)

            random_noise = np.random.randn(*affs.shape) * 0.001

            smoothed_affs = (
                gaussian_filter(affs, sigma=sigma) - 0.5
            ) * 0.01  # todo: parameterize?

            #######################

            shift = np.array(
                self.bias,
            ).reshape((-1, *((1,) * (len(affs.shape) - 1))))

            logger.error(
                f"unique seeds ({unique_seeds}) and seed counts: {np.unique(seeds, return_counts=True)}"
            )

            segs = mws.agglom(
                affs + shift + random_noise + smoothed_affs,
                offsets=self.affs_data.neighborhood,
                strides=self.strides,
                seeds=seeds,
                randomized_strides=self.randomized_strides,
            )

            logger.error(
                f"unique seeds ({unique_seeds}) and frag counts: {np.unique(segs, return_counts=True)}"
            )

            segs = segs * np.isin(segs, list(unique_seeds))

            logger.error(
                f"unique seeds ({unique_seeds}) and seg counts: {np.unique(segs, return_counts=True)}"
            )

            segs = Array(segs, block.read_roi.offset, segs_array.voxel_size)

            # store fragments
            segs_array[block.write_roi] = segs[block.write_roi]

            logger.info(f"releasing block: {block}")

        yield process_block



from contextlib import ExitStack
from itertools import product
from typing import TYPE_CHECKING

import daisy
import networkx as nx

if TYPE_CHECKING:
    from .blockwise import BlockwiseTask


class Pipeline:
    """
    A class to manage combinations of `BlockwiseTask`s that are grouped
    together in a pipeline.
    """

    task_graph: nx.DiGraph

    def __init__(self, task: "BlockwiseTask | None" = None):
        self.task_graph = nx.DiGraph()
        if task is not None:
            self.task_graph.add_node(task)

    def __add__(self, other: "Pipeline | BlockwiseTask") -> "Pipeline":
        """
        The task or pipeline (`task`) gets run in series after `self`.

        This means that every node in `self` without outgoing edges
        gets an edge to all nodes in `task` without incoming edges.
        """
        from volara.blockwise import BlockwiseTask

        if isinstance(other, BlockwiseTask):
            other = Pipeline(other)
        sink_nodes = [
            task
            for task in self.task_graph.nodes()
            if self.task_graph.out_degree(task) == 0
        ]
        source_nodes = [
            task
            for task in other.task_graph.nodes()
            if other.task_graph.in_degree(task) == 0
        ]

        combined_pipeline = Pipeline()
        combined_graph = combined_pipeline.task_graph
        combined_graph.add_nodes_from(self.task_graph.nodes())
        combined_graph.add_edges_from(self.task_graph.edges())
        combined_graph.add_nodes_from(other.task_graph.nodes())
        combined_graph.add_edges_from(other.task_graph.edges())
        for sink, source in product(sink_nodes, source_nodes):
            combined_graph.add_edge(sink, source)

        return combined_pipeline

    def __or__(self, other: "Pipeline | BlockwiseTask") -> "Pipeline":
        """
        The task or pipeline (`task`) gets run in parallel with `self`.

        Task graphs are merged, but no edges are added.
        """
        from volara.blockwise import BlockwiseTask

        if isinstance(other, BlockwiseTask):
            other = Pipeline(other)

        combined_pipeline = Pipeline()
        combined_graph = combined_pipeline.task_graph
        combined_graph.add_nodes_from(self.task_graph.nodes())
        combined_graph.add_edges_from(self.task_graph.edges())
        combined_graph.add_nodes_from(other.task_graph.nodes())
        combined_graph.add_edges_from(other.task_graph.edges())

        return combined_pipeline

    def run_blockwise(self, multiprocessing: bool = True):
        with ExitStack() as stack:
            node_ordering: list[BlockwiseTask] = list(
                nx.topological_sort(self.task_graph)
            )

            task_map: dict[BlockwiseTask, daisy.Task] = {}
            sinks = []
            for node in node_ordering:
                upstream_tasks = [
                    task_map[upstream]
                    for upstream in self.task_graph.predecessors(node)
                ]
                task = node.task(
                    upstream_tasks=upstream_tasks, multiprocessing=multiprocessing
                )
                task = stack.enter_context(task)
                task_map[node] = task
                if self.task_graph.out_degree(node) == 0:
                    sinks.append(task)

            all_tasks = list(task_map.values())

            if multiprocessing:
                daisy.run_blockwise(all_tasks)
            else:
                server = daisy.SerialServer()
                cl_monitor = daisy.cl_monitor.CLMonitor(server)  # noqa
                server.run_blockwise(all_tasks)

    def drop(self):
        for task in self.task_graph.nodes():
            task.drop()



from contextlib import contextmanager
from typing import Annotated, Literal

import daisy
import mwatershed as mws
import numpy as np
from funlib.geometry import Coordinate, Roi
from pydantic import Field

from volara.lut import LUT

from ..dbs import PostgreSQL, SQLite
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask

DB = Annotated[
    PostgreSQL | SQLite,
    Field(discriminator="db_type"),
]


class GraphMWS(BlockwiseTask):
    """
    Graph based execution of the MWS algorithm.
    Currently only supports executing in memory. The full graph for the given ROI
    is read into memory and then we run the mutex watershed algorithm on the full
    graph to get a globally optimal look up table.
    """

    task_type: Literal["graph-mws"] = "graph-mws"

    db: DB
    lut: LUT
    """
    The Look Up Table that will be saved on completion of this task.
    """
    starting_lut: LUT | None = None
    """
    An optional Look Up Table that provides a set of merged fragments that must
    be preserved in the final Look Up Table.
    """
    weights: dict[str, tuple[float, float]]
    """
    A dictionary of edge attributes and their weight and bias. These will be used
    to compute the edge weights for the mutex watershed algorithm. Positive edges
    will result in fragments merging, negative edges will result in splitting and
    edges will be processed in order of high to low magnitude.
    Each attribute will have a final score of `w * edge_data[attr] + b` for every
    `attr, (w, b) in weights.items()`
    If an attribute is not present in the edge data it will be skipped.
    """
    roi: tuple[PydanticCoordinate, PydanticCoordinate]
    """
    The roi to process. This is the roi of the full graph to process.
    """
    edge_per_attr: bool = True
    """
    Whether or not to create a separate edge for each attribute in the weights. If
    False, the sum of all the weighted attributes will be used as the only edge weight.
    """
    mean_attrs: dict[str, str] | None = None
    """
    A dictionary of attributes to compute the mean of for each segment. Given
    `mean_attrs = {"attr1": "out_attr1"}` and nodes `n_i` in a segment `s` we will
    set `s.out_attr1 = sum(n_i.attr1 * n_i.size) / sum(n_i.size)`.
    """
    out_db: DB | None = None
    """
    The db in which to store segment nodes and their attributes. Must not be None
    if `mean_attrs` is not None.
    """
    bounded_read: bool = True
    """
    Reading from the db can be made more efficient by not doing a spatial query
    and assuming we want all nodes and edges. If you don't want to process a
    sub volume of the graph setting this to false will speed up the read.
    """

    fit: Literal["shrink"] = "shrink"
    read_write_conflict: Literal[False] = False

    @property
    def task_name(self) -> str:
        return f"{self.lut.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        return Roi(*self.roi)

    @property
    def write_size(self) -> Coordinate:
        return self.write_roi.shape

    @property
    def context_size(self) -> Coordinate:
        return Coordinate((0,) * self.write_size.dims)

    @property
    def num_voxels_in_block(self) -> int:
        # We currently can't process in blocks
        return 1

    def drop_artifacts(self):
        self.lut.drop()
        if self.out_db is not None:
            self.out_db.drop()

    @contextmanager
    def process_block_func(self):
        rag_provider = self.db.open("r+")

        if self.out_db is not None:
            out_rag_provider = self.out_db.open("w")

        if self.starting_lut is not None:
            starting_frags, starting_segs = self.starting_lut.load()
            starting_map = {
                in_frag: out_frag
                for in_frag, out_frag in zip(starting_frags, starting_segs)
            }
        else:
            starting_map = None

        def process_block(block: daisy.Block):
            read_roi = block.write_roi if self.bounded_read else None
            node_attrs = (
                ["size"] + list(self.mean_attrs.keys())
                if self.mean_attrs is not None
                else []
            )
            graph = rag_provider.read_graph(
                read_roi, node_attrs=node_attrs, edge_attrs=list(self.weights.keys())
            )

            edges = []

            for u, v, edge_attrs in graph.edges(data=True):
                scores = [
                    w * edge_attrs.get(attr, None) + b
                    for attr, (w, b) in self.weights.items()
                    if edge_attrs.get(attr, None) is not None
                ]
                if self.edge_per_attr:
                    for score in scores:
                        edges.append((score, u, v))
                else:
                    edges.append((sum(scores), u, v))

            prefix_edges = []
            if starting_map is not None:
                groups: dict[int, set[int]] = {}
                for node in graph.nodes:
                    groups.setdefault(starting_map[node], set()).add(node)
                for group in groups.values():
                    pre_merged_ids = list(group)
                    for u, v in zip(pre_merged_ids, pre_merged_ids[1:]):
                        prefix_edges.append((True, u, v))

            edges = sorted(
                edges,
                key=lambda edge: abs(edge[0]),
                reverse=True,
            )
            edges = [(bool(aff > 0), u, v) for aff, u, v in edges]

            # generate the look up table via mutex watershed clustering
            mws_lut: list[tuple[int, int]] = mws.cluster(prefix_edges + edges)
            inputs: list[int]
            outputs: list[int]
            if len(mws_lut) > 0:
                inputs, outputs = [list(x) for x in zip(*mws_lut)]
            else:
                inputs, outputs = [], []

            lut = np.array([inputs, outputs])
            self.lut.save(lut, edges=edges)

            if self.mean_attrs is not None:
                assert self.out_db is not None, self.out_db
                out_graph = out_rag_provider.read_graph(block.write_roi)
                assert out_graph.number_of_nodes() == 0, out_graph.number_of_nodes
                mapping: dict[int, set[int]] = {}
                for in_frag, out_frag in zip(inputs, outputs):
                    in_group = mapping.setdefault(out_frag, set())
                    in_group.add(in_frag)

                for out_frag, in_group in mapping.items():
                    computed_codes = {
                        "seg_size": sum(
                            [graph.nodes[in_frag]["size"] for in_frag in in_group]
                        )
                    }
                    for in_code, out_code in self.mean_attrs.items():
                        out_codes = [
                            np.array(graph.nodes[in_frag][in_code])
                            * graph.nodes[in_frag]["size"]
                            for in_frag in in_group
                        ]
                        out_data = (
                            np.sum(np.array(out_codes), axis=0)
                            / computed_codes["seg_size"]
                        )
                        computed_codes[out_code] = out_data
                    for in_frag in in_group:
                        frag_attrs = graph.nodes[in_frag]
                        frag_attrs.update(computed_codes)
                        out_graph.add_node(in_frag, **frag_attrs)

                out_rag_provider.write_graph(out_graph, block.write_roi)

        yield process_block



import logging
import subprocess as sp
from abc import ABC
from pathlib import Path

import daisy

from .utils import StrictBaseModel

logger = logging.getLogger(__name__)


class Worker(StrictBaseModel, ABC):
    queue: str | None = None
    num_gpus: int = 0
    num_cpus: int = 1

    def get_command(self, config_path: Path, task_name: str) -> list[str]:
        cmd = [
            "volara-cli",
            "blockwise-worker",
            "-c",
            str(config_path),
        ]
        return cmd


class SlurmWorker(Worker):
    queue: str
    num_gpus: int = 0
    num_cpus: int = 1

    def get_command(self, config_path: Path, task_name: str) -> list[str]:
        cmd = super().get_command(config_path, task_name)

        context = daisy.Context.from_env()
        worker_id = context["worker_id"]
        task_id = context["task_id"]

        worker_log_basename = daisy.get_worker_log_basename(worker_id, task_id)

        log_file = worker_log_basename / "slurm_worker.log"
        log_error = worker_log_basename / "slurm_worker.err"

        return self.get_slurm_command(
            command=" ".join(cmd),
            execute=False,
            expand=False,
            queue=self.queue,
            num_gpus=self.num_gpus,
            num_cpus=self.num_cpus,
            log_file=log_file,
            error_file=log_error,
        )

    def is_sbatch_available(self) -> bool:
        try:
            _result = sp.run(
                ["sbatch", "--version"], capture_output=True, text=True, check=True
            )
            # successful, return True
            return True
        except sp.CalledProcessError as e:
            # errors in the subprocess
            raise RuntimeError(f"sbatch failed to execute: {e}") from e
        except FileNotFoundError:
            # sbatch is not found in the system's PATH
            raise EnvironmentError(
                "sbatch is not installed or not in PATH. Either install sbatch on your cluster, or run locally."
            )

    def get_slurm_command(
        self,
        command: str,
        num_cpus: int = 1,
        num_gpus: int = 0,
        memory: int = 15564,
        constraint: str = "",
        queue: str = "",
        execute: bool = False,
        expand: bool = True,
        job_name: str = "",
        array_size: int = 1,
        array_limit: int | None = None,
        log_file: str | None = None,
        error_file: str | None = None,
        flags: list[str] | None = None,
    ) -> list[str]:
        """
        Prepares and optionally executes a command on a slurm cluster,

        Args:
            command (str): The command to be executed within the Slurm job.
            num_cpus (int, optional): Number of CPU cores per task. Defaults to 1.
            num_gpus (int, optional): Number of GPUs required. Defaults to 0.
            memory (int, optional): Memory allocation (in MB) for the job. Defaults
                to 25600.
            constraint (str, optional): Constraint specification for job
                execution. Defaults to "".
            queue (str, optional): Name of the Slurm partition (queue) to submit the
                job. Defaults to "".
            execute (bool, optional): Whether to execute the command or just return
                the command. Defaults to False.
            expand (bool, optional): Returns a string if True, and a list if False.
                Defaults to True.
            job_name (str, optional): Name assigned to the Slurm job. Defaults to "".
            array_size (int, optional): If greater than 1, submits a job array of
                this size. Defaults to 1.
            array_limit (int | None, optional): Limits the number of
                simultaneously running tasks in the job array. Defaults to None.
            log_file (str | None, optional): Path for standard output logging.
                Defaults to None.
            error_file (str | None, optional): Path for standard error logging.
                Defaults to None.
            flags (list[str] | None, optional): Additional sbatch flags as a
                list. Defaults to None.

        Returns:
            str | list[str] | None: Depending on `execute` and `expand`,
                returns the job ID (if executed), the constructed command as a string or
                list, or None if an error occurred.
        """

        # TODO: raises exception on failure. Maybe handle this gracefully?
        self.is_sbatch_available()

        if execute:
            logging.info(
                f"Scheduling job on {num_cpus} CPUs, {num_gpus} GPUs with {memory} MB on queue {queue}"
            )

        log = f"--output={log_file}" if log_file else "--output=%x_%j.log"
        error = f"--error={error_file}" if error_file else "--error=%x_%j.err"

        use_gpus = f"--gpus={num_gpus}" if num_gpus > 0 else ""
        use_constraint = (
            f"--constraint={constraint}" if constraint and constraint != "None" else ""
        )
        job_name_cmd = f"--job-name={job_name}" if job_name else ""

        # Initialize the command list with the base sbatch command
        run_command = ["sbatch"]

        # Job array handling
        if array_size > 1:
            array_cmd = (
                f"--array=1-{array_size}{f'%{array_limit}' if array_limit else ''}"
            )
            run_command.append(array_cmd)

        # Append job name, CPU, GPU, memory, queue, and constraint constraints to the command
        if job_name_cmd:
            run_command.append(job_name_cmd)
        run_command.append(f"--cpus-per-task={num_cpus}")
        if use_gpus:
            run_command.append(use_gpus)
        run_command.append(f"--mem={memory}")
        if queue:
            run_command.append(f"--partition={queue}")
        if use_constraint:
            run_command.append(use_constraint)

        # Append log and error file paths to the command
        run_command.append(log)
        run_command.append(error)

        # Append additional flags if provided
        if flags:
            run_command.extend(flags)

        # Append the command to be executed within the job
        run_command.append(f"--wrap={command}")

        return run_command


class LSFWorker(Worker):
    queue: str
    num_gpus: int = 0
    num_cpus: int = 1

    def get_command(self, config_path: Path, task_name: str) -> list[str]:
        cmd = super().get_command(config_path, task_name)

        context = daisy.Context.from_env()
        worker_id = context["worker_id"]
        task_id = context["task_id"]

        worker_log_basename = daisy.get_worker_log_basename(worker_id, task_id)
        if not worker_log_basename.exists():
            worker_log_basename.mkdir(parents=True, exist_ok=True)

        log_file = worker_log_basename / "lsf_worker.log"
        log_error = worker_log_basename / "lsf_worker.err"

        return self.get_lsf_command(
            command=cmd,
            queue=self.queue,
            num_cpus=self.num_cpus,
            num_gpus=self.num_gpus,
            log_file=log_file,
            error_file=log_error,
        )

    def is_bsub_available(self) -> bool:
        try:
            _result = sp.run(["bsub", "-V"], capture_output=True, text=True, check=True)
            # successful, return True
            return True
        except sp.CalledProcessError as e:
            # errors in the subprocess
            raise RuntimeError(f"bsub failed to execute: {e}") from e
        except FileNotFoundError:
            # bsub is not found in the system's PATH
            raise EnvironmentError(
                "bsub is not installed or not in PATH. Either install bsub on your cluster, or run locally."
            )

    def get_lsf_command(
        self,
        command: list[str],
        num_cpus: int = 1,
        num_gpus: int = 0,
        queue: str = "",
        log_file: str | None = None,
        error_file: str | None = None,
    ) -> list[str]:
        """
        Prepares and optionally executes a command on an LSF cluster,

        Args:
            command (str): The command to be executed within the LSF job.
            num_cpus (int, optional): Number of CPU cores per task. Defaults to 1.
            num_gpus (int, optional): Number of GPUs required. Defaults to 0.
            queue (str, optional): Name of the LSF queue to submit the job.
                Defaults to "".
            log_file (str | None, optional): Path for standard output logging.
                Defaults to None.
            error_file (str | None, optional): Path for standard error logging.
                Defaults to None.
        """
        self.is_bsub_available()

        log = ["-o", str(log_file)] if log_file is not None else []
        error = ["-e", str(error_file)] if error_file is not None else []

        run_command = ["bsub"]

        run_command.extend(["-n", str(num_cpus)])
        if num_gpus > 0:
            run_command.extend(["-num-gpus", str(num_gpus)])
        if queue:
            run_command.extend(["-q", str(queue)])

        run_command.extend(log)
        run_command.extend(error)

        run_command += command

        return run_command


class LocalWorker(Worker):
    def get_command(self, config_path: Path, task_name: str) -> list[str]:
        cmd = super().get_command(config_path, task_name)

        context = daisy.Context.from_env()
        worker_id = context["worker_id"]
        task_id = context["task_id"]

        worker_log_basename = daisy.get_worker_log_basename(worker_id, task_id)

        _log_file = worker_log_basename / "out.log"
        _log_error = worker_log_basename / "out.err"

        # TODO: update command to use log files, test that they exist
        # current tests only show that "worker_id" and "task_id" can be retrieved
        return cmd



from importlib.metadata import entry_points
from typing import Annotated, Union

from pydantic import Field, TypeAdapter

from .aff_agglom import AffAgglom as AffAgglom
from .argmax import Argmax as Argmax
from .blockwise import BlockwiseTask
from .distance_agglom import (
    DistanceAgglom as DistanceAgglom,
)
from .extract_frags import ExtractFrags as ExtractFrags
from .graph_mws import GraphMWS as GraphMWS
from .register import ApplyShift as ApplyShift
from .register import ComputeShift as ComputeShift
from .relabel import Relabel as Relabel
from .seeded_extract_frags import SeededExtractFrags as SeededExtractFrags
from .threshold import Threshold as Threshold

BLOCKWISE_TASKS = []


def register_task(task: BlockwiseTask):
    if task not in BLOCKWISE_TASKS:
        BLOCKWISE_TASKS.append(task)


def discover_tasks():
    for entry_point in entry_points().select(group="volara.blockwise_tasks"):
        task_class = entry_point.load()
        register_task(task_class)


def get_task(task_type: str) -> BlockwiseTask:
    """
    Get the task class for a given task type. This is useful for dynamically fetching
    tasks from the base `volara.blockwise` module, despite some tasks potentially
    being defined in other modules as plugins via entry points. This allows `volara`
    to serialize and execute external tasks in isolated environments such as on
    cluster workers.
    """
    for task in BLOCKWISE_TASKS:
        parsed_task_type = task.model_fields["task_type"].default
        if parsed_task_type == task_type:
            return task
    raise ValueError(f"Unknown task: {task_type}, {BLOCKWISE_TASKS}")


TASKS_DISCOVERED = False


def get_blockwise_tasks_type():
    global TASKS_DISCOVERED
    if not TASKS_DISCOVERED:
        discover_tasks()
        TASKS_DISCOVERED = True
    return TypeAdapter(
        Annotated[
            Union[tuple(BLOCKWISE_TASKS)],
            Field(discriminator="task_type"),
        ]
    )


__all__ = [
    "BlockwiseTask",
    "AffAgglom",
    "ApplyShift",
    "Argmax",
    "ComputeShift",
    "DistanceAgglom",
    "ExtractFrags",
    "GraphMWS",
    "Relabel",
    "SeededExtractFrags",
    "Threshold",
]



from contextlib import contextmanager
from shutil import rmtree
from typing import Literal

import numpy as np
from funlib.geometry import Coordinate, Roi

from volara.lut import LUT
from volara.tmp import replace_values

from ..datasets import Dataset, Labels
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask


class Relabel(BlockwiseTask):
    """
    A task for blockwise relabelling of arrays using a lookup table from fragment
    to segment IDs.
    """

    task_type: Literal["relabel"] = "relabel"
    frags_data: Labels
    """
    The fragments dataset from which we read the fragment IDs.
    """
    seg_data: Labels
    """
    The segments dataset to which we write the relabeled segment IDs.
    """
    lut: LUT
    """
    The path to the lookup table (LUT) that maps fragment IDs to segment IDs.
    """
    block_size: PydanticCoordinate

    fit: Literal["shrink"] = "shrink"
    read_write_conflict: Literal[False] = False
    _out_array_dtype: np.dtype = np.dtype(np.uint64)

    @property
    def task_name(self) -> str:
        return f"{self.seg_data.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.frags_data.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def voxel_size(self) -> Coordinate:
        return self.frags_data.array("r").voxel_size

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.voxel_size

    @property
    def context_size(self) -> Coordinate:
        return Coordinate((0,) * self.write_size.dims)

    @property
    def output_datasets(self) -> list[Dataset]:
        return [self.seg_data]

    def drop_artifacts(self):
        rmtree(self.seg_data.store, ignore_errors=True)

    def init(self):
        self.init_out_array()

    def init_out_array(self):
        in_data = self.frags_data.array("r")
        self.seg_data.prepare(
            self.write_roi.shape / self.voxel_size,
            self.write_size / self.voxel_size,
            self.write_roi.offset,
            self.voxel_size,
            units=in_data.units,
            axis_names=in_data.axis_names,
            types=in_data.types,
            dtype=self._out_array_dtype,
        )

    def map_block(self, block, frags, segs, mapping):
        segs[block.write_roi] = replace_values(
            frags.to_ndarray(block.write_roi), mapping[0], mapping[1]
        )

    @contextmanager
    def process_block_func(self):
        frags = self.frags_data.array("r")
        segs = self.seg_data.array("r+")

        def process_block(block):
            mapping = self.lut.load()
            self.map_block(block, frags, segs, mapping)

        yield process_block



import logging
from contextlib import contextmanager
from itertools import chain, combinations
from typing import Annotated, Literal

import numpy as np
from funlib.geometry import Coordinate, Roi
from pydantic import Field
from scipy.ndimage import laplace
from scipy.spatial import cKDTree

from volara.lut import LUT

from ..datasets import Dataset, Labels
from ..dbs import PostgreSQL, SQLite
from ..utils import PydanticCoordinate
from .blockwise import BlockwiseTask

logger = logging.getLogger(__file__)


class DistanceAgglom(BlockwiseTask):
    """
    Distance based edge creation for fragment to segment agglomeration.
    In this context distance is defined as the distance between the embeddings
    of two fragments. Given some attributes saved or computable on the fragments
    we compute the distance between pairs of fragments and then save the edges
    with their costs.
    Because the distance is computed between embeddings, there is no spatial limit
    to the distance between fragments. This means that without a spatial constraint
    we would need to compute the distance between all pairs of fragments. To avoid
    this we can set a connection radius that will limit the euclidean distance between
    connectable fragments.
    """

    task_type: Literal["distance-agglom"] = "distance-agglom"

    storage: (
        Annotated[
            PostgreSQL | SQLite,
            Field(discriminator="db_type"),
        ]
        | LUT
    )
    """
    Where to store the edges and or the final Look Up Table.
    If a Database is provided, it is assumed that each fragment in the
    fragments dataset has a node in the graph already saved.
    """
    frags_data: Labels
    """
    The labels dataset that contains the fragments to agglomerate
    """
    distance_keys: list[str] | None = None
    background_intensities: list[float] | None = None
    eps: float = 1e-8
    distance_threshold: float | None = None
    distance_metric: Literal["euclidean", "cosine", "max"] = "cosine"

    block_size: PydanticCoordinate
    context: PydanticCoordinate

    fit: Literal["shrink"] = "shrink"
    read_write_conflict: Literal[False] = False

    @property
    def task_name(self) -> str:
        return f"{self.frags_data.name}-{self.task_type}"

    @property
    def write_roi(self) -> Roi:
        total_roi = self.frags_data.array("r").roi
        if self.roi is not None:
            total_roi = total_roi.intersect(Roi(self.roi[0], self.roi[1]))
        return total_roi

    @property
    def write_size(self) -> Coordinate:
        return self.block_size * self.frags_data.array("r").voxel_size

    @property
    def context_size(self) -> Coordinate:
        return self.context * self.frags_data.array("r").voxel_size

    @property
    def output_datasets(self) -> list[Dataset]:
        return []

    def drop_artifacts(self):
        self.db.drop_edges()

    def label_distances(self, labels, voxel_size, dist_threshold=0.0):
        # First 0 out all voxel where the laplace is 0 (not an edge voxel)
        output = np.zeros_like(labels, dtype=np.float32)
        object_filter = laplace(labels, output=output)
        labels *= abs(object_filter) > 0
        coordinates = np.nonzero(labels)
        labels = labels[*coordinates]
        coords = np.column_stack(coordinates) * np.array(voxel_size)

        trees = []
        for label in np.unique(labels):
            trees.append(
                (label, cKDTree(coords[labels == label]), coords[labels == label])
            )
        min_dists = {}

        for (label_a, tree_a, tree_coords_a), (
            label_b,
            tree_b,
            tree_coords_b,
        ) in combinations(trees, 2):
            pairs = tree_a.query_ball_tree(tree_b, dist_threshold)
            if len(list(chain(*pairs))) > 0:
                min_dists[(label_a, label_b)] = min(
                    np.linalg.norm(tree_coords_a[i] - tree_coords_b[j])
                    for i, matches in enumerate(pairs)
                    for j in matches
                )

        return list(min_dists.keys()), list(min_dists.values())

    def agglomerate_in_block(self, block, frags, rag_provider):
        voxel_size = frags.voxel_size
        frags = frags.to_ndarray(block.read_roi, fill_value=0)
        rag = rag_provider[block.read_roi]

        distance_threshold = (
            self.distance_threshold
            if self.distance_threshold is not None
            else min(self.context * voxel_size)
        )
        pairs, distances = self.label_distances(frags, voxel_size, distance_threshold)
        distance_keys = [] if self.distance_keys is None else self.distance_keys
        background_intensities = (
            [0.0] * len(distance_keys)
            if self.background_intensities is None
            else self.background_intensities
        )
        assert len(background_intensities) == len(distance_keys)
        for (frag_i, frag_j), dist in zip(pairs, distances):
            if frag_i in rag.nodes and frag_j in rag.nodes:
                node_attrs_a = rag.nodes[frag_i]
                node_attrs_b = rag.nodes[frag_j]
                attr_dict = {"distance": dist}
                for distance_key, background_intensity in zip(
                    distance_keys, background_intensities
                ):
                    if (
                        distance_key not in node_attrs_a
                        or distance_key not in node_attrs_b
                    ):
                        continue
                    distance_a = (
                        np.array(node_attrs_a[distance_key]) - background_intensity
                    )
                    distance_b = (
                        np.array(node_attrs_b[distance_key]) - background_intensity
                    )
                    if self.distance_metric == "cosine":
                        similarity = np.dot(distance_a, distance_b) / max(
                            np.linalg.norm(distance_a) * np.linalg.norm(distance_b),
                            self.eps,
                        )
                    elif self.distance_metric == "euclidean":
                        similarity = -np.linalg.norm(distance_a - distance_b)
                    elif self.distance_metric == "max":
                        similarity = -np.max(np.abs(distance_a - distance_b))
                    attr_dict[f"{distance_key}_similarity"] = similarity
                rag.add_edge(
                    int(frag_i),
                    int(frag_j),
                    **attr_dict,
                )

        rag_provider.write_graph(rag, block.write_roi, write_nodes=False)

    @contextmanager
    def process_block_func(self):
        frags = self.frags_data.array("r")
        rag_provider = self.db.open("r+")

        def process_block(block):
            self.agglomerate_in_block(
                block,
                frags,
                rag_provider,
            )

        yield process_block



from funlib.geometry import Coordinate
from pydantic import BaseModel, ConfigDict
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema


class PydanticCoordinate(Coordinate):
    """A Pydantic-compatible Coordinate type."""

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type, _handler):
        return core_schema.chain_schema(
            [
                core_schema.list_schema(
                    items_schema=core_schema.int_schema(), min_length=1
                ),
                core_schema.no_info_plain_validator_function(cls.validate),
            ]
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema, _handler) -> JsonSchemaValue:
        """Defines the JSON schema for OpenAPI compatibility."""
        return {"type": "array", "items": {"type": "integer"}, "minItems": 1}

    @classmethod
    def validate(cls, v):
        if isinstance(v, Coordinate):
            return v  # Already a Coordinate
        if isinstance(v, (tuple, list)) and all(isinstance(i, int) for i in v):
            return Coordinate(v)  # Convert tuple/list to Coordinate
        raise ValueError(f"Invalid coordinate: {v}")


class StrictBaseModel(BaseModel):
    """
    A BaseModel that does not allow for extra fields.
    """

    model_config = ConfigDict(extra="forbid")
